"""
 @file
 @brief Shared AI Tools context-menu builder for project files and timeline.
"""

import os
from functools import partial

from PyQt5.QtGui import QIcon

from classes.app import get_app
from classes import info
from .menu import StyledContextMenu


def _trigger_generation(win, template_id, source_file=None, open_dialog=False):
    win.generation_service.action_generate_trigger(
        source_file=source_file,
        template_id=template_id,
        open_dialog=open_dialog,
    )


def _icon(name):
    icon_path = os.path.join(info.PATH, "themes", "cosmic", "images", name)
    if os.path.exists(icon_path):
        return QIcon(icon_path)
    return QIcon()


def add_ai_tools_menu(win, parent_menu, source_file=None):
    _ = get_app()._tr
    if not win.is_comfy_available(force=False):
        return None

    grouped = win.generation_service.build_menu_templates(source_file=source_file)
    menu_defs = []
    if source_file:
        menu_defs = [("enhance", _("Enhance with AI")), ("unknown", _("Unknown AI"))]
    else:
        menu_defs = [("create", _("Create with AI")), ("unknown", _("Unknown AI"))]

    created_menus = []
    for key, title in menu_defs:
        templates = list(grouped.get(key, []) or [])
        if not templates:
            continue
        ai_menu = StyledContextMenu(title=title, parent=parent_menu)
        ai_menu.setIcon(_icon("tool-generate-sparkle.svg"))

        parent_labels = {
            "track_object": _("Track an Object"),
            "extract": _("Extract"),
        }
        parent_icons = {
            "track_object": "tool-generate-sparkle.svg",
            "extract": "ai-action-create-image.svg",
        }
        parent_menus = {}

        inserted_style_separator = False
        for template in templates:
            template_key = str(template.get("template_id") or template.get("id") or "")
            if (
                key == "enhance"
                and not inserted_style_separator
                and template_key in ("img2img-basic", "video2video-basic")
            ):
                ai_menu.addSeparator()
                inserted_style_separator = True

            open_dialog = template.get("open_dialog")
            if not isinstance(open_dialog, bool):
                open_dialog = (source_file is None) or bool(template.get("needs_prompt", False))

            template_parent = str(template.get("menu_parent") or "").strip().lower()
            target_menu = ai_menu
            if key == "enhance" and template_parent:
                if template_parent not in parent_menus:
                    submenu_title = parent_labels.get(template_parent, template_parent.replace("_", " ").title())
                    submenu = StyledContextMenu(title=submenu_title, parent=ai_menu)
                    submenu.setIcon(_icon(parent_icons.get(template_parent, "tool-generate-sparkle.svg")))
                    ai_menu.addMenu(submenu)
                    parent_menus[template_parent] = submenu
                target_menu = parent_menus[template_parent]

            action = target_menu.addAction(_(str(template.get("display_name", ""))))
            action.setIcon(_icon(win.generation_service.icon_for_template(template)))
            action.triggered.connect(
                partial(
                    _trigger_generation,
                    win,
                    template.get("id"),
                    source_file,
                    open_dialog,
                )
            )

        parent_menu.addMenu(ai_menu)
        created_menus.append(ai_menu)

    return created_menus[0] if created_menus else None



"""
 @file
 @brief Transition geometry helpers for the timeline widget.
 @author Jonathan Thomas <jonathan@openshot.org>

 @section LICENSE

 Copyright (c) 2008-2025 OpenShot Studios, LLC
 (http://www.openshotstudios.com). This file is part of
 OpenShot Video Editor (http://www.openshot.org), an open-source project
 dedicated to delivering high quality video editing and animation solutions
 to the world.

 OpenShot Video Editor is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 OpenShot Video Editor is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with OpenShot Library.  If not, see <http://www.gnu.org/licenses/>.
 """

from PyQt5.QtCore import QRectF

from classes.query import Transition

from .base import _GeometryEntry


class TransitionGeometryMixin:
    """Populate cached transition rectangles."""

    def _populate_transition_rects(self, layers, ctx, win):
        w = self.widget
        overrides_map = getattr(w, "_pending_transition_overrides", {})
        entries = []
        selected_ids = {
            str(item_id)
            for item_id in (getattr(win, "selected_transitions", []) or [])
        }
        for tr in Transition.filter():
            tr_data = tr.data if isinstance(tr.data, dict) else {}
            override = overrides_map.get(tr.id, {})

            position = override.get("position", tr_data.get("position", 0.0))
            start = override.get("start", tr_data.get("start", 0.0))
            end = override.get("end", tr_data.get("end", start))
            layer_val = override.get("layer", tr_data.get("layer", 0))

            try:
                position = float(position)
            except (TypeError, ValueError):
                position = 0.0
            try:
                start = float(start)
            except (TypeError, ValueError):
                start = 0.0
            try:
                end = float(end)
            except (TypeError, ValueError):
                end = start
            if end < start:
                end = start
            try:
                layer_key = int(layer_val)
            except (TypeError, ValueError):
                layer_key = layer_val

            tx = w.track_name_width + position * w.pixels_per_second
            layer_idx = layers.get(layer_key, 0)
            offset = ctx.get("track_offsets", {}).get(
                w.normalize_track_number(layer_key),
                layer_idx * ctx["spacing"],
            )
            ty = w.ruler_height + ctx.get("top_margin", 0.0) + offset
            tw = (end - start) * w.pixels_per_second
            rect = QRectF(tx, ty, tw, w.vertical_factor)
            entries.append((rect.left(), rect, tr))

        def _transition_sort_key(entry):
            left, rect, tran = entry
            return left, rect.x(), getattr(tran, "id", "")

        entries.sort(key=_transition_sort_key)
        transition_entries = []
        transition_starts = []
        max_right = float("-inf")
        max_rights = []
        for left, rect, tran in entries:
            is_selected = str(getattr(tran, "id", "")) in selected_ids
            transition_entries.append(
                _GeometryEntry(rect=rect, obj=tran, selected=is_selected)
            )
            transition_starts.append(left)
            max_right = max(max_right, rect.right())
            max_rights.append(max_right)
        self.transition_entries = transition_entries
        self._transition_starts = transition_starts
        self._transition_max_rights = max_rights



"""
 @file
 @brief This file contains Comfy generation orchestration logic.
 @author Jonathan Thomas <jonathan@openshot.org>

 @section LICENSE

 Copyright (c) 2008-2026 OpenShot Studios, LLC
 (http://www.openshotstudios.com). This file is part of
 OpenShot Video Editor (http://www.openshot.org), an open-source project
 dedicated to delivering high quality video editing and animation solutions
 to the world.

 OpenShot Video Editor is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 OpenShot Video Editor is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with OpenShot Library.  If not, see <http://www.gnu.org/licenses/>.
 """

import os
import re
import tempfile
import json
import random
from time import time
from urllib.parse import unquote
from fractions import Fraction

import openshot
from PyQt5.QtCore import QObject, QThread, pyqtSignal, pyqtSlot
from PyQt5.QtWidgets import QMessageBox, QDialog

from classes import info
from classes import time_parts
from classes.app import get_app
from classes.comfy_client import ComfyClient
from classes.comfy_templates import ComfyTemplateRegistry
from classes.logger import log
from classes.query import File
from windows.generate import GenerateMediaDialog


RASTER_IMAGE_EXTENSIONS = {
    ".png", ".jpg", ".jpeg", ".webp", ".bmp", ".tif", ".tiff", ".gif",
}


def is_supported_img2img_path(path):
    path_text = str(path or "").strip()
    # Comfy annotated paths can look like: "image.jpg [input]"
    if path_text.endswith("]") and " [" in path_text:
        path_text = path_text.rsplit(" [", 1)[0].strip()
    ext = os.path.splitext(path_text)[1].lower()
    return ext in RASTER_IMAGE_EXTENSIONS


class _ComfyAvailabilityWorker(QObject):
    finished = pyqtSignal(int, str, bool, str)

    @pyqtSlot(int, str, float)
    def check(self, request_id, url, timeout):
        available = False
        error_text = ""
        try:
            available = ComfyClient(url).ping(timeout=float(timeout))
        except Exception as ex:
            error_text = str(ex)
        self.finished.emit(int(request_id), str(url), bool(available), str(error_text or ""))


class GenerationService(QObject):
    """Encapsulates generation-specific UI + workflow behavior."""
    SAM2_DEFAULT_TARGET_BATCH_BYTES = 4 * 1024 * 1024 * 1024  # 4 GiB
    SAM2_ESTIMATED_BYTES_PER_PIXEL = 24.0
    SAM2_ESTIMATED_BYTES_PER_PIXEL_HIGHLIGHT = 64.0
    SAM2_ESTIMATED_BYTES_PER_PIXEL_BLUR = 40.0
    SAM2_MIN_FRAMES_PER_BATCH = 4
    SAM2_MAX_FRAMES_PER_BATCH = 192
    _request_comfy_check = pyqtSignal(int, str, float)

    def __init__(self, win):
        super().__init__(win)
        self.win = win
        self._generation_temp_files = []
        self._comfy_status_cache = {"checked_at": 0.0, "available": False, "url": ""}
        self._last_logged_comfy_state = None
        self.template_registry = ComfyTemplateRegistry()
        self._comfy_check_request_id = 0
        self._latest_comfy_check_request_id = 0
        self._comfy_check_callbacks = {}
        self._comfy_check_thread = QThread(self)
        self._comfy_check_thread.setObjectName("comfy_availability_worker")
        self._comfy_check_worker = _ComfyAvailabilityWorker()
        self._comfy_check_worker.moveToThread(self._comfy_check_thread)
        self._request_comfy_check.connect(self._comfy_check_worker.check)
        self._comfy_check_worker.finished.connect(self._on_comfy_check_finished)
        self._comfy_check_thread.start()

    def cleanup_temp_files(self):
        for tmp_path in list(self._generation_temp_files):
            try:
                if tmp_path and os.path.exists(tmp_path):
                    os.remove(tmp_path)
            except OSError:
                pass
        self._generation_temp_files = []

    def comfy_ui_url(self):
        url = get_app().get_settings().get("comfy-ui-url") or "http://127.0.0.1:8188"
        return str(url).strip().rstrip("/")

    def shutdown(self):
        if getattr(self, "_comfy_check_thread", None):
            self._comfy_check_thread.quit()
            self._comfy_check_thread.wait()

    def refresh_comfy_availability_async(self, timeout=0.5, callback=None):
        self._comfy_check_request_id += 1
        request_id = self._comfy_check_request_id
        url = self.comfy_ui_url()
        self._latest_comfy_check_request_id = request_id
        if callable(callback):
            self._comfy_check_callbacks[request_id] = callback
        self._request_comfy_check.emit(request_id, url, float(timeout))
        return request_id

    def is_comfy_available(self, force=False):
        url = self.comfy_ui_url()
        if not force:
            if str(self._comfy_status_cache.get("url", "")) != url:
                return False
            return self._comfy_status_cache["available"]

        now = time()
        available = False
        error_text = ""
        try:
            available = ComfyClient(url).ping(timeout=0.5)
        except Exception as ex:
            available = False
            error_text = str(ex)

        previous_available = bool(self._comfy_status_cache.get("available"))
        previous_url = str(self._comfy_status_cache.get("url", ""))
        self._comfy_status_cache["checked_at"] = now
        self._comfy_status_cache["available"] = available
        self._comfy_status_cache["url"] = url

        state = (url, bool(available))
        if force or state != self._last_logged_comfy_state or previous_url != url or previous_available != available:
            if available:
                log.info("ComfyUI check passed at %s", url)
            else:
                if error_text:
                    log.info("ComfyUI check failed at %s (%s)", url, error_text)
                else:
                    log.info("ComfyUI check failed at %s", url)
            self._last_logged_comfy_state = state
        return available

    @pyqtSlot(int, str, bool, str)
    def _on_comfy_check_finished(self, request_id, url, available, error_text):
        callback = self._comfy_check_callbacks.pop(int(request_id), None)
        if int(request_id) != int(self._latest_comfy_check_request_id):
            return

        previous_available = bool(self._comfy_status_cache.get("available"))
        previous_url = str(self._comfy_status_cache.get("url", ""))
        self._comfy_status_cache["checked_at"] = time()
        self._comfy_status_cache["available"] = bool(available)
        self._comfy_status_cache["url"] = str(url)

        state = (str(url), bool(available))
        if state != self._last_logged_comfy_state or previous_url != str(url) or previous_available != bool(available):
            if available:
                log.info("ComfyUI check passed at %s", url)
            else:
                if error_text:
                    log.info("ComfyUI check failed at %s (%s)", url, error_text)
                else:
                    log.info("ComfyUI check failed at %s", url)
            self._last_logged_comfy_state = state

        if callback:
            try:
                callback(bool(available), str(error_text or ""), str(url))
            except Exception:
                log.debug("ComfyUI availability callback failed", exc_info=1)

    def can_open_generate_dialog(self):
        return len(self.win.selected_file_ids()) <= 1

    def _prepare_generation_source_path(self, source_file, template_id):
        if not source_file:
            return ""

        source_path = source_file.data.get("path", "")
        media_type = source_file.data.get("media_type")
        if template_id not in ("img2img-basic", "upscale-realesrgan-x4", "img2video-wan") or media_type != "image":
            return source_path

        if is_supported_img2img_path(source_path):
            return source_path

        tmp_fd, tmp_png = tempfile.mkstemp(prefix="openshot-comfy-", suffix=".png")
        os.close(tmp_fd)
        try:
            clip = openshot.Clip(source_path)
            frame = clip.Reader().GetFrame(1)
            frame.Save(tmp_png, 1.0)
            self._generation_temp_files.append(tmp_png)
            return tmp_png
        except Exception:
            try:
                os.remove(tmp_png)
            except OSError:
                pass
            raise

    def _default_generation_name(self, source_file):
        default_name = "generation"
        if source_file:
            path = source_file.data.get("path", "")
            if path:
                default_name = "{}_gen".format(os.path.splitext(os.path.basename(path))[0])
        return default_name

    def _get_source_dimensions(self, source_file):
        if not source_file:
            return (0, 0)
        data = source_file.data if hasattr(source_file, "data") and isinstance(source_file.data, dict) else {}
        try:
            width = int(data.get("width", 0) or 0)
        except Exception:
            width = 0
        try:
            height = int(data.get("height", 0) or 0)
        except Exception:
            height = 0
        return (max(0, width), max(0, height))

    def _sam2_target_batch_bytes(self):
        settings = get_app().get_settings()
        raw_bytes = settings.get("comfy-sam2-target-batch-bytes")
        if raw_bytes is not None:
            try:
                value = int(raw_bytes)
                if value > 0:
                    return value
            except Exception:
                pass
        raw_gb = settings.get("comfy-sam2-target-batch-gb")
        if raw_gb is not None:
            try:
                value = float(raw_gb)
                if value > 0.0:
                    return int(value * 1024 * 1024 * 1024)
            except Exception:
                pass
        return int(self.SAM2_DEFAULT_TARGET_BATCH_BYTES)

    def _estimate_sam2_frames_per_batch(self, width, height, bytes_per_pixel=None):
        width = int(max(0, width))
        height = int(max(0, height))
        if width <= 0 or height <= 0:
            return self.SAM2_MIN_FRAMES_PER_BATCH
        target_bytes = self._sam2_target_batch_bytes()
        if bytes_per_pixel is None:
            bytes_per_pixel = self.SAM2_ESTIMATED_BYTES_PER_PIXEL
        try:
            bytes_per_pixel = float(bytes_per_pixel)
        except Exception:
            bytes_per_pixel = float(self.SAM2_ESTIMATED_BYTES_PER_PIXEL)
        bytes_per_frame = max(
            1.0,
            float(width) * float(height) * bytes_per_pixel,
        )
        frames = int(target_bytes / bytes_per_frame)
        frames = max(self.SAM2_MIN_FRAMES_PER_BATCH, min(self.SAM2_MAX_FRAMES_PER_BATCH, frames))
        # Keep chunk sizes aligned for more stable batching behavior.
        frames = max(self.SAM2_MIN_FRAMES_PER_BATCH, int((frames // 4) * 4))
        return frames

    def _apply_dynamic_sam2_meta_batch(self, workflow, source_file, template_id=None):
        template_id = str(template_id or "").strip().lower()
        # Only adjust SAM2 video tracking template workflows.
        if template_id and template_id not in (
            "video-blur-anything-sam2",
            "video-highlight-anything-sam2",
            "video-mask-anything-sam2",
        ):
            return
        if not isinstance(workflow, dict):
            return

        width, height = self._get_source_dimensions(source_file)
        if width <= 0 or height <= 0:
            return

        has_sam2_chunked = False
        for node in workflow.values():
            if not isinstance(node, dict):
                continue
            class_type = str(node.get("class_type", "")).strip().lower()
            if class_type == "openshotsam2videosegmentationchunked":
                has_sam2_chunked = True
                break
        if not has_sam2_chunked:
            return

        # Account for downstream per-frame processing memory:
        # - Highlight path is the heaviest (multiple full-frame tensor intermediates)
        # - Blur path is moderately heavy
        # - Mask-only path is closest to baseline SAM2 estimate
        estimated_bpp = float(self.SAM2_ESTIMATED_BYTES_PER_PIXEL)
        for node in workflow.values():
            if not isinstance(node, dict):
                continue
            class_type = str(node.get("class_type", "")).strip().lower()
            if class_type == "openshotimagehighlightmasked":
                estimated_bpp = max(estimated_bpp, float(self.SAM2_ESTIMATED_BYTES_PER_PIXEL_HIGHLIGHT))
            elif class_type == "openshotimageblurmasked":
                estimated_bpp = max(estimated_bpp, float(self.SAM2_ESTIMATED_BYTES_PER_PIXEL_BLUR))

        dynamic_frames = self._estimate_sam2_frames_per_batch(width, height, bytes_per_pixel=estimated_bpp)
        updated_chunk_nodes = 0
        updated_batch_nodes = 0
        for node in workflow.values():
            if not isinstance(node, dict):
                continue
            class_type = str(node.get("class_type", "")).strip().lower()
            inputs = node.get("inputs", {})
            if not isinstance(inputs, dict):
                continue
            if class_type == "openshotsam2videosegmentationchunked" and "chunk_size_frames" in inputs:
                inputs["chunk_size_frames"] = int(dynamic_frames)
                updated_chunk_nodes += 1
            if class_type == "vhs_batchmanager" and "frames_per_batch" in inputs:
                inputs["frames_per_batch"] = int(dynamic_frames)
                updated_batch_nodes += 1
        if updated_chunk_nodes or updated_batch_nodes:
            log.info(
                "Dynamic SAM2 batch size: %s frames (source=%sx%s, target_bytes=%s, est_bpp=%s, template=%s, chunk_nodes=%s, batch_nodes=%s)",
                dynamic_frames,
                width,
                height,
                self._sam2_target_batch_bytes(),
                round(estimated_bpp, 2),
                template_id or "unknown",
                updated_chunk_nodes,
                updated_batch_nodes,
            )

    def templates_for_context(self, source_file=None):
        templates = self.template_registry.templates_for_context(source_file=source_file)
        return [
            {"id": t.get("id"), "name": t.get("display_name"), "template": t}
            for t in templates
        ]

    def build_menu_templates(self, source_file=None):
        grouped = {"create": [], "enhance": [], "unknown": []}
        for template in self.template_registry.templates_for_context(source_file=source_file):
            category = str(template.get("category", "unknown"))
            if category not in grouped:
                category = "unknown"
            grouped[category].append(template)
        return grouped

    def icon_for_template(self, template):
        return self.template_registry.output_icon_name(template)

    def _prepare_template_workflow(
        self,
        template,
        payload_name,
        prompt_text,
        source_file,
        source_path,
        reference_image_path="",
        coordinates_positive_text="",
        coordinates_negative_text="",
        rectangles_positive_text="",
        rectangles_negative_text="",
        auto_mode=False,
        tracking_selection=None,
        highlight_color="",
        highlight_opacity=0.0,
        border_color="",
        border_width=0,
        mask_brightness=1.0,
        background_brightness=1.0,
    ):
        workflow = self.template_registry.get_workflow_copy(template.get("id"))
        if not workflow:
            raise ValueError("Template workflow not found.")

        template_dir = ""
        template_path = str((template or {}).get("path") or "").strip()
        if template_path:
            template_dir = os.path.dirname(template_path)

        def _resolve_template_local_file(path_text):
            path_text = str(path_text or "").strip()
            if not path_text:
                return ""
            if os.path.isabs(path_text):
                return path_text if os.path.exists(path_text) else ""
            if not template_dir:
                return ""
            candidate = os.path.abspath(os.path.join(template_dir, path_text))
            if os.path.exists(candidate):
                return candidate
            return ""

        template_id = str((template or {}).get("id") or "").strip().lower()
        prompt_text = str(prompt_text or "").strip()
        music_prompt_text = prompt_text
        music_lyrics_text = ""
        if template_id == "txt2music-ace-step" and prompt_text:
            # Optional inline format:
            #   <style/tags text>
            #   Lyrics:
            #   <lyrics lines...>
            split_match = re.split(r"(?im)^\s*lyrics\s*:\s*", prompt_text, maxsplit=1)
            if len(split_match) == 2:
                music_prompt_text = str(split_match[0] or "").strip()
                music_lyrics_text = str(split_match[1] or "").strip()
            if not music_prompt_text:
                music_prompt_text = prompt_text
        coordinates_positive_text = str(coordinates_positive_text or "").strip()
        coordinates_negative_text = str(coordinates_negative_text or "").strip()
        rectangles_positive_text = str(rectangles_positive_text or "").strip()
        rectangles_negative_text = str(rectangles_negative_text or "").strip()
        auto_mode = bool(auto_mode)
        tracking_selection = tracking_selection if isinstance(tracking_selection, dict) else {}
        highlight_color = str(highlight_color or "").strip()
        border_color = str(border_color or "").strip()
        try:
            highlight_opacity = float(highlight_opacity)
        except (TypeError, ValueError):
            highlight_opacity = 0.0
        try:
            border_width = int(border_width)
        except (TypeError, ValueError):
            border_width = 0
        try:
            mask_brightness = float(mask_brightness)
        except (TypeError, ValueError):
            mask_brightness = 1.0
        try:
            background_brightness = float(background_brightness)
        except (TypeError, ValueError):
            background_brightness = 1.0
        media_type = str(source_file.data.get("media_type", "")).strip().lower() if source_file else ""
        run_seed_base = random.randint(1, 2**31 - 1)
        seed_counter = 0
        applied_prompt = False
        loadimage_node_ids = []
        loadvideo_node_ids = []
        loadaudio_node_ids = []

        for node_id, node in workflow.items():
            if not isinstance(node, dict):
                continue
            class_flat = str(node.get("class_type", "")).strip().lower()
            if class_flat == "loadimage":
                loadimage_node_ids.append(str(node_id))
            elif class_flat in ("loadvideo", "load video", "vhs_loadvideo", "vhs_loadvideopath", "vhs_loadvideoffmpegpath"):
                loadvideo_node_ids.append(str(node_id))
            elif class_flat in ("loadaudio", "load audio"):
                loadaudio_node_ids.append(str(node_id))

        def _is_placeholder_value(path_text):
            path_text = str(path_text or "").strip().lower()
            return path_text in ("__openshot_input__", "{{openshot_input}}", "$openshot_input")

        def _is_reference_image_placeholder_value(path_text):
            path_text = str(path_text or "").strip().lower()
            return path_text in (
                "__openshot_reference_image__",
                "{{openshot_reference_image}}",
                "$openshot_reference_image",
            )

        def _is_prompt_placeholder_value(text_value):
            text_value = str(text_value or "").strip().lower()
            return text_value in ("__openshot_prompt__", "{{openshot_prompt}}", "$openshot_prompt")

        def _is_lyrics_placeholder_value(text_value):
            text_value = str(text_value or "").strip().lower()
            return text_value in ("__openshot_lyrics__", "{{openshot_lyrics}}", "$openshot_lyrics")

        def _replace_prompt_placeholders(text_value):
            text_value = str(text_value or "")
            if not text_value:
                return text_value
            return (
                text_value
                .replace("__openshot_prompt__", prompt_text)
                .replace("{{openshot_prompt}}", prompt_text)
                .replace("$openshot_prompt", prompt_text)
            )

        def _replace_lyrics_placeholders(text_value):
            text_value = str(text_value or "")
            if not text_value:
                return text_value
            return (
                text_value
                .replace("__openshot_lyrics__", music_lyrics_text)
                .replace("{{openshot_lyrics}}", music_lyrics_text)
                .replace("$openshot_lyrics", music_lyrics_text)
            )

        def _normalize_sam2_coords_input(text_value, fallback_value):
            text_value = str(text_value or "").strip()
            fallback_value = str(fallback_value or "").strip()
            if not text_value:
                return fallback_value
            # Accept raw JSON list format expected by Sam2VideoSegmentationAddPoints.
            if text_value.startswith("[") and "x" in text_value and "y" in text_value:
                return text_value

            # Also accept "x,y; x,y" shorthand and convert to JSON list.
            points = []
            for chunk in text_value.split(";"):
                chunk = chunk.strip()
                if not chunk:
                    continue
                parts = [p.strip() for p in chunk.split(",")]
                if len(parts) != 2:
                    return fallback_value
                try:
                    x_val = float(parts[0])
                    y_val = float(parts[1])
                except (TypeError, ValueError):
                    return fallback_value
                points.append({"x": x_val, "y": y_val})
            if not points:
                return fallback_value
            return str(points).replace("'", "\"")

        def _next_random_seed(existing_value):
            nonlocal seed_counter
            seed_counter += 1
            try:
                seed_offset = int(existing_value)
            except Exception:
                seed_offset = 0
            return int((run_seed_base + seed_offset + (seed_counter * 7919)) % (2**31 - 1)) or 1

        def _parse_sam2_points(coords_text):
            coords_text = str(coords_text or "").strip()
            if not coords_text:
                return []
            try:
                parsed = json.loads(coords_text.replace("'", "\""))
            except Exception:
                return []
            if not isinstance(parsed, list):
                return []
            points = []
            for item in parsed:
                if not isinstance(item, dict):
                    continue
                if "x" not in item or "y" not in item:
                    continue
                try:
                    points.append({"x": float(item["x"]), "y": float(item["y"])})
                except Exception:
                    continue
            return points

        def _select_bind_nodes(node_ids, path_keys, preferred_upload=None):
            if isinstance(path_keys, str):
                path_keys = [path_keys]
            explicit = []
            candidates = []
            for node_id in node_ids:
                node = workflow.get(node_id, {})
                inputs = node.get("inputs", {}) if isinstance(node, dict) else {}
                if not isinstance(inputs, dict):
                    continue
                path_value = ""
                for path_key in path_keys:
                    candidate_value = str(inputs.get(path_key, "")).strip()
                    if candidate_value:
                        path_value = candidate_value
                        break
                upload_value = str(inputs.get("upload", "")).strip().lower()
                if _is_placeholder_value(path_value):
                    explicit.append(node_id)
                    continue
                if preferred_upload and upload_value == preferred_upload:
                    explicit.append(node_id)
                    continue
                if not path_value:
                    candidates.append(node_id)
            if explicit:
                return set(explicit)
            if candidates:
                return {candidates[0]}
            if node_ids:
                return {node_ids[0]}
            return set()

        image_bind_nodes = _select_bind_nodes(loadimage_node_ids, ["image"], preferred_upload="image")
        video_bind_nodes = _select_bind_nodes(loadvideo_node_ids, ["file", "video"])
        audio_bind_nodes = _select_bind_nodes(loadaudio_node_ids, ["audio", "file"])

        for node_id, node in workflow.items():
            if not isinstance(node, dict):
                continue
            class_type = str(node.get("class_type", "")).strip()
            inputs = node.get("inputs", {})
            if not isinstance(inputs, dict):
                continue

            class_flat = class_type.lower().strip()

            # Resolve generic OpenShot source placeholders in any string input
            # (custom nodes may use keys like `video_path` instead of `video`/`file`).
            if source_path or reference_image_path:
                for input_key, input_value in list(inputs.items()):
                    if isinstance(input_value, str) and source_path and _is_placeholder_value(input_value):
                        inputs[input_key] = source_path
                    elif isinstance(input_value, str) and reference_image_path and _is_reference_image_placeholder_value(input_value):
                        inputs[input_key] = reference_image_path

            if "filename_prefix" in inputs:
                prefix_value = str(inputs.get("filename_prefix", "")).strip()
                if "/" in prefix_value:
                    head, tail = prefix_value.rsplit("/", 1)
                    tail = str(tail or "output").strip()
                    inputs["filename_prefix"] = "{}/{}_{}".format(head, tail, payload_name)
                else:
                    inputs["filename_prefix"] = payload_name

            if "seed" in inputs and not isinstance(inputs.get("seed", None), (list, dict)):
                inputs["seed"] = _next_random_seed(inputs.get("seed"))

            if prompt_text:
                text_value = inputs.get("text", None)
                prompt_value = inputs.get("prompt", None)
                tags_value = inputs.get("tags", None)
                lyrics_value = inputs.get("lyrics", None)

                if isinstance(text_value, str):
                    replaced_text = _replace_prompt_placeholders(text_value)
                    if replaced_text != text_value:
                        inputs["text"] = replaced_text
                        applied_prompt = True
                        text_value = replaced_text
                if isinstance(prompt_value, str):
                    replaced_prompt = _replace_prompt_placeholders(prompt_value)
                    if replaced_prompt != prompt_value:
                        inputs["prompt"] = replaced_prompt
                        applied_prompt = True
                        prompt_value = replaced_prompt
                if isinstance(tags_value, str):
                    replaced_tags = _replace_prompt_placeholders(tags_value)
                    if replaced_tags != tags_value:
                        inputs["tags"] = replaced_tags
                        applied_prompt = True
                        tags_value = replaced_tags

                if isinstance(text_value, str) and _is_prompt_placeholder_value(text_value):
                    inputs["text"] = prompt_text
                    applied_prompt = True
                elif isinstance(prompt_value, str) and _is_prompt_placeholder_value(prompt_value):
                    inputs["prompt"] = prompt_text
                    applied_prompt = True
                elif isinstance(tags_value, str) and _is_prompt_placeholder_value(tags_value):
                    # Support music nodes that use "tags" instead of "text"/"prompt".
                    if template_id == "txt2music-ace-step":
                        inputs["tags"] = music_prompt_text
                    else:
                        inputs["tags"] = prompt_text
                    applied_prompt = True
                elif isinstance(lyrics_value, str) and _is_lyrics_placeholder_value(lyrics_value):
                    if template_id == "txt2music-ace-step":
                        inputs["lyrics"] = music_lyrics_text
                    else:
                        inputs["lyrics"] = ""
                elif class_flat == "cliptextencode" and "text" in inputs and not applied_prompt:
                    inputs["text"] = prompt_text
                    applied_prompt = True
                elif "prompt" in inputs and isinstance(prompt_value, str) and not prompt_value.strip() and not applied_prompt:
                    # Support prompt-driven custom nodes (e.g. GroundingDINO/SAM2) that expose a plain string prompt input.
                    inputs["prompt"] = prompt_text
                    applied_prompt = True

            if isinstance(inputs.get("lyrics", None), str):
                lyrics_value = inputs.get("lyrics", "")
                replaced_lyrics = _replace_lyrics_placeholders(lyrics_value)
                if replaced_lyrics != lyrics_value:
                    inputs["lyrics"] = replaced_lyrics

            if class_flat in (
                "sam2videosegmentationaddpoints",
                "sam2segmentation",
                "openshotsam2videosegmentationaddpoints",
                "openshotsam2segmentation",
            ):
                coords_text = coordinates_positive_text
                points = _parse_sam2_points(coords_text)
                has_positive_rects = bool(rectangles_positive_text)

                seed_frame_idx = 0
                if isinstance(tracking_selection, dict):
                    try:
                        seed_frame_idx = max(0, int(tracking_selection.get("seed_frame", 1)) - 1)
                    except Exception:
                        seed_frame_idx = 0
                if "frame_index" in inputs:
                    try:
                        inputs["frame_index"] = int(seed_frame_idx)
                    except Exception:
                        pass
                if "tracking_selection_json" in inputs:
                    try:
                        inputs["tracking_selection_json"] = json.dumps(tracking_selection or {})
                    except Exception:
                        inputs["tracking_selection_json"] = "{}"
                if "dino_prompt" in inputs:
                    inputs["dino_prompt"] = str(prompt_text or "")

                has_dino_prompt = bool(str(prompt_text or "").strip()) and ("dino_prompt" in inputs)

                auto_enabled = bool(inputs.get("auto_mode", False)) or auto_mode
                if "auto_mode" in inputs:
                    inputs["auto_mode"] = bool(auto_enabled)
                if ("anything-sam2" in template_id) and (not points) and (not has_positive_rects) and (not auto_enabled) and (not has_dino_prompt):
                    raise ValueError("No SAM2 seed was provided. Use Points, Rectangle, or Auto mode.")

                # New OpenShot node contract.
                if "positive_points_json" in inputs and isinstance(inputs.get("positive_points_json", None), str):
                    inputs["positive_points_json"] = _normalize_sam2_coords_input(
                        coords_text,
                        str(inputs.get("positive_points_json", "")),
                    )
                # Backward compatibility for third-party node variants.
                elif "coordinates_positive" in inputs and isinstance(inputs.get("coordinates_positive", None), str):
                    inputs["coordinates_positive"] = _normalize_sam2_coords_input(
                        coords_text,
                        str(inputs.get("coordinates_positive", "")),
                    )

                if class_flat in ("sam2segmentation", "openshotsam2segmentation") and "individual_objects" in inputs:
                    # For Blur Anything, treat points as a single combined prompt.
                    # This is more stable with mixed positive/negative points and avoids
                    # per-object mask selection quirks in the current SAM2 single-image node.
                    if "blur-anything-sam2" in template_id:
                        inputs["individual_objects"] = False
                    else:
                        # Non-Blur-Anything templates keep multi-object behavior.
                        inputs["individual_objects"] = bool(len(points) > 1)

                if coordinates_negative_text:
                    if "negative_points_json" in inputs and isinstance(inputs.get("negative_points_json", None), str):
                        inputs["negative_points_json"] = _normalize_sam2_coords_input(
                            coordinates_negative_text,
                            str(inputs.get("negative_points_json", "")),
                        )
                    elif "coordinates_negative" in inputs:
                        neg_value = inputs.get("coordinates_negative", "")
                        if isinstance(neg_value, str) or "coordinates_negative" not in inputs:
                            inputs["coordinates_negative"] = _normalize_sam2_coords_input(
                                coordinates_negative_text,
                                str(neg_value or ""),
                            )
                if rectangles_positive_text and ("positive_rects_json" in inputs) and isinstance(
                    inputs.get("positive_rects_json", None), str
                ):
                    inputs["positive_rects_json"] = rectangles_positive_text
                if rectangles_negative_text and ("negative_rects_json" in inputs) and isinstance(
                    inputs.get("negative_rects_json", None), str
                ):
                    inputs["negative_rects_json"] = rectangles_negative_text

            if class_flat in ("openshotimagehighlightmasked",) or class_type.strip() == "OpenShotImageHighlightMasked":
                if "highlight_color" in inputs and highlight_color:
                    inputs["highlight_color"] = highlight_color
                if "highlight_opacity" in inputs:
                    inputs["highlight_opacity"] = float(max(0.0, min(1.0, highlight_opacity)))
                if "border_color" in inputs and border_color:
                    inputs["border_color"] = border_color
                if "border_width" in inputs:
                    inputs["border_width"] = int(max(0, border_width))
                if "mask_brightness" in inputs:
                    inputs["mask_brightness"] = float(max(0.0, min(3.0, mask_brightness)))
                if "background_brightness" in inputs:
                    inputs["background_brightness"] = float(max(0.0, min(3.0, background_brightness)))

            if not source_path:
                continue

            node_id = str(node_id)

            if class_flat == "loadimage" and media_type == "image" and node_id in image_bind_nodes:
                if "image" in inputs:
                    inputs["image"] = source_path
                if "upload" in inputs:
                    inputs["upload"] = "image"
            elif class_flat == "loadimage":
                # Resolve template-local reference images (relative filenames) to absolute paths,
                # so ComfyClient can upload and rewrite them to [input] automatically.
                configured_image = str(inputs.get("image", "")).strip()
                local_image = _resolve_template_local_file(configured_image)
                if local_image:
                    inputs["image"] = local_image
                    if "upload" in inputs:
                        inputs["upload"] = "image"
                elif media_type == "image" and source_path:
                    # If additional reference images are missing, gracefully fallback to the selected source image.
                    # This keeps common exported workflows usable without hand-editing filenames.
                    missing_relative = configured_image and (not os.path.isabs(configured_image))
                    missing_absolute = os.path.isabs(configured_image) and (not os.path.exists(configured_image))
                    if missing_relative or missing_absolute:
                        inputs["image"] = source_path
                        if "upload" in inputs:
                            inputs["upload"] = "image"
                        log.warning(
                            "Comfy template missing LoadImage asset (%s). Falling back to selected source image.",
                            configured_image,
                        )
            elif class_flat in ("loadvideo", "load video", "vhs_loadvideopath", "vhs_loadvideoffmpegpath") and media_type == "video" and node_id in video_bind_nodes:
                if "file" in inputs:
                    inputs["file"] = source_path
                elif "video" in inputs:
                    inputs["video"] = source_path
            elif class_flat in ("loadvideo", "load video", "vhs_loadvideopath", "vhs_loadvideoffmpegpath"):
                local_video = _resolve_template_local_file(inputs.get("file", ""))
                if local_video and "file" in inputs:
                    inputs["file"] = local_video
            elif class_flat == "vhs_loadvideo" and media_type == "video" and node_id in video_bind_nodes:
                if "video" in inputs:
                    inputs["video"] = source_path
            elif class_flat in ("loadaudio", "load audio") and media_type == "audio" and node_id in audio_bind_nodes:
                if "audio" in inputs:
                    inputs["audio"] = source_path
                elif "file" in inputs:
                    inputs["file"] = source_path
            elif class_flat in ("loadaudio", "load audio"):
                local_audio = _resolve_template_local_file(inputs.get("audio", "") or inputs.get("file", ""))
                if local_audio:
                    if "audio" in inputs:
                        inputs["audio"] = local_audio
                    elif "file" in inputs:
                        inputs["file"] = local_audio

        if media_type == "image" and image_bind_nodes:
            log.debug("Comfy template image input binding nodes=%s source=%s", sorted(image_bind_nodes), source_path)
        if media_type == "video" and video_bind_nodes:
            log.debug("Comfy template video input binding nodes=%s source=%s", sorted(video_bind_nodes), source_path)
        if media_type == "audio" and audio_bind_nodes:
            log.debug("Comfy template audio input binding nodes=%s source=%s", sorted(audio_bind_nodes), source_path)

        self._apply_dynamic_sam2_meta_batch(workflow, source_file=source_file, template_id=template_id)

        return workflow

    def _save_nodes_for_workflow(self, workflow, template_id=None):
        template_id = str(template_id or "").strip().lower()
        save_nodes = []
        for node_id, node in workflow.items():
            if not isinstance(node, dict):
                continue
            class_type = str(node.get("class_type", "")).strip().lower()
            if not class_type:
                continue
            if class_type.startswith("save") or class_type in (
                "previewany",
                "transnetv2_run",
                "openshottransnetscenedetect",
                "vhs_videocombine",
            ):
                if class_type == "vhs_videocombine":
                    inputs = node.get("inputs", {}) if isinstance(node.get("inputs", {}), dict) else {}
                    prefix = str(inputs.get("filename_prefix", "")).strip().lower()
                    is_mask_output = ("openshot_mask" in prefix)
                    is_track_template = template_id in (
                        "video-blur-anything-sam2",
                        "video-highlight-anything-sam2",
                        "video-mask-anything-sam2",
                    )
                    if is_track_template:
                        if template_id == "video-mask-anything-sam2":
                            if not is_mask_output:
                                continue
                        else:
                            if is_mask_output:
                                continue
                save_nodes.append(str(node_id))
        return save_nodes

    def action_generate_trigger(self, checked=True, source_file=None, template_id=None, open_dialog=True):
        selected_files = [source_file] if source_file else self.win.selected_files()
        if len(selected_files) > 1:
            return

        if not self.is_comfy_available(force=True):
            msg = QMessageBox(self.win)
            msg.setWindowTitle("ComfyUI Unavailable")
            msg.setText(
                "OpenShot could not connect to ComfyUI at:\n{}\n\n"
                "Start ComfyUI or update the URL in Preferences > Experimental.".format(self.comfy_ui_url())
            )
            msg.exec_()
            return

        source_file = selected_files[0] if selected_files else None
        templates = self.templates_for_context(source_file=source_file)
        available_template_ids = {str(t.get("id", "")).strip() for t in templates}
        if open_dialog:
            dialog_title = "Enhance with AI" if source_file else "Create with AI"
            win = GenerateMediaDialog(
                source_file=source_file,
                templates=templates,
                preselected_template_id=template_id,
                dialog_title=dialog_title,
                parent=self.win,
            )
            if win.exec_() != QDialog.Accepted:
                return
            payload = win.get_payload()
        else:
            selected_template_id = str(template_id or "").strip()
            if not selected_template_id:
                return
            if selected_template_id not in available_template_ids:
                QMessageBox.information(
                    self.win,
                    "Invalid Input",
                    "The selected AI action is not available for this source type.",
                )
                return
            payload = {
                "name": self._default_generation_name(source_file),
                "template_id": selected_template_id,
                "prompt": "",
            }

        payload_name = self._next_generation_name(payload.get("name"))
        source_file_id = source_file.id if source_file else None
        template_meta = self.template_registry.get_template(payload.get("template_id"))
        if not template_meta:
            QMessageBox.information(self.win, "Invalid Input", "The selected AI template was not found.")
            return
        try:
            source_path = self._prepare_generation_source_path(source_file, payload.get("template_id"))
        except Exception as ex:
            QMessageBox.warning(
                self.win,
                "Source Conversion Failed",
                "OpenShot could not convert this image into PNG for ComfyUI.\n\n{}".format(ex),
            )
            return
        reference_image_path = ""
        reference_image_file_id = str(payload.get("reference_image_file_id") or "").strip()
        if reference_image_file_id:
            reference_image_file = File.get(id=reference_image_file_id)
            if reference_image_file and isinstance(reference_image_file.data, dict):
                reference_image_path = str(reference_image_file.data.get("path", "") or "").strip()
        try:
            workflow = self._prepare_template_workflow(
                template_meta,
                payload_name=payload_name,
                prompt_text=payload.get("prompt"),
                source_file=source_file,
                source_path=source_path,
                reference_image_path=reference_image_path,
                coordinates_positive_text=payload.get("coordinates_positive"),
                coordinates_negative_text=payload.get("coordinates_negative"),
                rectangles_positive_text=payload.get("rectangles_positive"),
                rectangles_negative_text=payload.get("rectangles_negative"),
                auto_mode=payload.get("auto_mode"),
                tracking_selection=payload.get("tracking_selection"),
                highlight_color=payload.get("highlight_color"),
                highlight_opacity=payload.get("highlight_opacity"),
                border_color=payload.get("border_color"),
                border_width=payload.get("border_width"),
                mask_brightness=payload.get("mask_brightness"),
                background_brightness=payload.get("background_brightness"),
            )
        except Exception as ex:
            QMessageBox.information(self.win, "Invalid Input", str(ex))
            return
        request = {
            "comfy_url": self.comfy_ui_url(),
            "workflow": workflow,
            "client_id": "openshot-qt",
            "timeout_s": 21600,
            "save_node_ids": self._save_nodes_for_workflow(workflow, template_id=payload.get("template_id")),
            "template_id": str(payload.get("template_id") or ""),
        }
        job_id = self.win.generation_queue.enqueue(
            payload_name,
            payload.get("template_id"),
            payload.get("prompt"),
            source_file_id=source_file_id,
            request=request,
        )
        if not job_id:
            QMessageBox.information(
                self.win,
                "Generation Already Active",
                "Only one active generation is allowed per source file.",
            )
            return

        self.win.statusBar.showMessage("Queued generation job", 3000)

    def on_generation_job_finished(self, job_id, status):
        job = self.win.generation_queue.get_job(job_id) if getattr(self.win, "generation_queue", None) else None
        if not job:
            return

        if status == "completed":
            result = self._import_generation_outputs(job)
            imported = int(result.get("imported", 0))
            caption_saved = bool(result.get("caption_saved", False))
            scenes_labeled = int(result.get("scenes_labeled", 0))
            scene_splits_created = int(result.get("scene_splits_created", 0))
            if imported > 0 and caption_saved:
                self.win.statusBar.showMessage(
                    "Generation completed, imported {} file(s), and saved file caption data".format(imported),
                    5000,
                )
            elif scene_splits_created > 0:
                self.win.statusBar.showMessage(
                    "Generation completed and created {} scene split file(s)".format(scene_splits_created),
                    5000,
                )
            elif imported > 0 and scenes_labeled > 0:
                self.win.statusBar.showMessage(
                    "Generation completed, imported {} file(s), and labeled {} scene segment(s)".format(
                        imported, scenes_labeled
                    ),
                    5000,
                )
            elif imported > 0:
                self.win.statusBar.showMessage("Generation completed and imported {} file(s)".format(imported), 5000)
            elif caption_saved:
                self.win.statusBar.showMessage("Generation completed and saved file caption data", 5000)
            else:
                self.win.statusBar.showMessage("Generation completed (no output files found)", 5000)
            return

        if status == "canceled":
            self.win.statusBar.showMessage("Generation canceled", 3000)
            return

        if status == "failed":
            error_text = ComfyClient.summarize_error_text(job.get("error") or "ComfyUI generation failed.")
            self.win.statusBar.showMessage("Generation failed", 5000)
            QMessageBox.warning(self.win, "Generation Failed", error_text)

    def _import_generation_outputs(self, job):
        outputs = list(job.get("outputs", []) or [])
        if not outputs:
            return {"imported": 0, "caption_saved": False, "scene_splits_created": 0}

        request = job.get("request", {}) or {}
        comfy_url = str(request.get("comfy_url") or self.comfy_ui_url())
        client = ComfyClient(comfy_url)
        output_dir = info.COMFYUI_OUTPUT_PATH
        os.makedirs(output_dir, exist_ok=True)
        template_id = str(job.get("template_id") or "").strip().lower()

        name_raw = str(job.get("name") or "generation")
        safe_name = re.sub(r"[^A-Za-z0-9._-]+", "_", name_raw).strip("._")
        if not safe_name:
            safe_name = "generation"

        saved_paths = []
        text_outputs = []
        scene_splits_created = 0
        video_path_exts = {".mp4", ".mov", ".mkv", ".webm", ".avi", ".m4v"}
        seen_video_payload_paths = set()
        for index, output_ref in enumerate(outputs, start=1):
            text_payload = str(output_ref.get("text", "")).strip()
            if text_payload:
                if template_id == "video-segment-scenes-transnet":
                    scene_ranges = self._extract_scene_ranges_from_text(text_payload)
                    if scene_ranges:
                        created = self._create_scene_split_files(
                            source_file_id=job.get("source_file_id"),
                            scene_ranges=scene_ranges,
                        )
                        scene_splits_created += created
                        if created > 0:
                            continue
                payload_video_paths = self._extract_video_paths_from_text(text_payload)
                if not payload_video_paths:
                    payload_ext = os.path.splitext(text_payload)[1].lower()
                    if payload_ext in video_path_exts:
                        payload_video_paths = [text_payload]
                downloaded_any_video = False
                for raw_video_path in payload_video_paths:
                    norm_video_path = str(raw_video_path).strip().replace("\\", "/")
                    if not norm_video_path or norm_video_path in seen_video_payload_paths:
                        continue
                    seen_video_payload_paths.add(norm_video_path)

                    payload_ext = os.path.splitext(norm_video_path)[1].lower() or ".mp4"
                    video_ref = self._comfy_output_ref_from_path(norm_video_path)
                    if not video_ref:
                        continue
                    local_name = "{}_{}{}".format(safe_name, str(index).zfill(3), payload_ext)
                    local_path = self._next_available_path(os.path.join(output_dir, local_name))
                    try:
                        client.download_output_file(video_ref, local_path)
                        saved_paths.append(local_path)
                        downloaded_any_video = True
                    except Exception as ex:
                        log.warning(
                            "Failed to download segmented video from Comfy path output %s: %s",
                            raw_video_path,
                            ex,
                        )
                if downloaded_any_video:
                    continue

                # Some Save SRT node variants return the output file path as text.
                # Convert that path to a downloadable Comfy output ref when possible.
                if text_payload.lower().endswith(".srt"):
                    srt_ref = self._comfy_output_ref_from_path(text_payload)
                    if srt_ref:
                        local_name = "{}_{}{}".format(safe_name, str(index).zfill(3), ".srt")
                        local_path = self._next_available_path(os.path.join(output_dir, local_name))
                        try:
                            client.download_output_file(srt_ref, local_path)
                            with open(local_path, "r", encoding="utf-8") as handle:
                                srt_text = handle.read().strip()
                            if srt_text:
                                saved_paths.append(local_path)
                                text_outputs.append(srt_text)
                                continue
                        except Exception as ex:
                            log.warning("Failed to download/read SRT from Comfy path output %s: %s", text_payload, ex)
                    if self._looks_like_filesystem_path(text_payload):
                        # Some Whisper/SRT nodes return only a server-side output path.
                        # Don't treat that path string as caption text.
                        continue

                ext = ".srt" if str(output_ref.get("format", "")).lower() == "srt" else ".txt"
                local_name = "{}_{}{}".format(safe_name, str(index).zfill(3), ext)
                local_path = self._next_available_path(os.path.join(output_dir, local_name))
                try:
                    with open(local_path, "w", encoding="utf-8") as handle:
                        handle.write(text_payload)
                    saved_paths.append(local_path)
                    text_outputs.append(text_payload)
                except Exception as ex:
                    log.warning("Failed to write Comfy text output to %s: %s", local_path, ex)
                continue

            original_name = str(output_ref.get("filename", "output.png"))
            ext = os.path.splitext(original_name)[1] or ".png"
            local_name = "{}_{}{}".format(safe_name, str(index).zfill(3), ext)
            local_path = self._next_available_path(os.path.join(output_dir, local_name))
            try:
                client.download_output_file(output_ref, local_path)
                saved_paths.append(local_path)
            except Exception as ex:
                log.warning("Failed to download Comfy output %s: %s", output_ref, ex)

        if saved_paths:
            self.win.files_model.add_files(
                saved_paths,
                quiet=True,
                prevent_image_seq=True,
                prevent_recent_folder=True,
            )

        if not saved_paths and scene_splits_created <= 0:
            return {"imported": 0, "caption_saved": False, "scene_splits_created": 0}

        caption_saved = False
        scenes_labeled = 0
        if template_id == "video-whisper-srt":
            caption_text = self._resolve_caption_text(saved_paths, text_outputs)
            caption_saved = self._store_caption_on_file(
                source_file_id=job.get("source_file_id"),
                caption_text=caption_text,
            )
        if template_id == "video-segment-scenes-transnet" and saved_paths:
            scenes_labeled = self._apply_scene_segment_metadata(
                source_file_id=job.get("source_file_id"),
                saved_paths=saved_paths,
            )
        return {
            "imported": len(saved_paths),
            "caption_saved": caption_saved,
            "scenes_labeled": scenes_labeled,
            "scene_splits_created": scene_splits_created,
        }

    def _extract_video_paths_from_text(self, text_payload):
        """Extract absolute video file paths from log/text payloads."""
        text_payload = str(text_payload or "")
        if not text_payload:
            return []
        pattern = re.compile(
            r"([A-Za-z]:[\\/][^\r\n]+?\.(?:mp4|mov|mkv|webm|avi|m4v)|/[^\r\n]+?\.(?:mp4|mov|mkv|webm|avi|m4v))",
            re.IGNORECASE,
        )
        return [match.strip() for match in pattern.findall(text_payload) if match.strip()]

    def _looks_like_filesystem_path(self, text_payload):
        text_payload = str(text_payload or "").strip()
        if not text_payload:
            return False
        normalized = text_payload.replace("\\", "/")
        if re.match(r"^[A-Za-z]:/", normalized):
            return True
        return normalized.startswith("/")

    def _extract_scene_ranges_from_text(self, text_payload):
        """Parse scene range metadata JSON from text output payloads."""
        text_payload = str(text_payload or "").strip()
        if not text_payload:
            return []
        if not text_payload.startswith("{"):
            first = text_payload.find("{")
            last = text_payload.rfind("}")
            if first >= 0 and last > first:
                text_payload = text_payload[first:last + 1]
            else:
                return []
        try:
            payload = json.loads(text_payload)
        except Exception:
            return []

        segment_entries = payload.get("segments") if isinstance(payload, dict) else None
        if not isinstance(segment_entries, list):
            return []

        scene_ranges = []
        for segment in segment_entries:
            if not isinstance(segment, dict):
                continue
            start_seconds = segment.get("start_seconds", segment.get("start"))
            end_seconds = segment.get("end_seconds", segment.get("end"))
            try:
                start_value = float(start_seconds)
                end_value = float(end_seconds)
            except (TypeError, ValueError):
                continue
            if end_value <= start_value:
                continue
            scene_ranges.append((max(0.0, start_value), max(0.0, end_value)))
        return scene_ranges

    def _create_scene_split_files(self, source_file_id, scene_ranges):
        """Create split-style file entries pointing to the same source media path."""
        source_file = File.get(id=source_file_id) if source_file_id else None
        if source_file is None:
            source_file = File.get(id=str(source_file_id or ""))
        if source_file is None:
            return 0

        source_data = source_file.data if isinstance(source_file.data, dict) else {}
        source_path = str(source_data.get("path", "") or "")
        if not source_path:
            return 0
        if str(source_data.get("media_type", "")).lower() != "video":
            return 0

        fps_data = source_data.get("fps", {})
        fps_fraction = Fraction(30, 1)
        try:
            fps_num = int(fps_data.get("num", 30))
            fps_den = int(fps_data.get("den", 1) or 1)
            if fps_num > 0 and fps_den > 0:
                fps_fraction = Fraction(fps_num, fps_den)
        except (TypeError, ValueError, ZeroDivisionError):
            fps_fraction = Fraction(30, 1)

        source_duration = float(source_data.get("duration") or 0.0)
        base_name = os.path.splitext(os.path.basename(source_path))[0] or "scene"

        created = 0
        for start_seconds, end_seconds in scene_ranges:
            start_seconds = max(0.0, float(start_seconds or 0.0))
            end_seconds = max(start_seconds, float(end_seconds or start_seconds))
            if source_duration > 0.0:
                start_seconds = min(start_seconds, source_duration)
                end_seconds = min(end_seconds, source_duration)
            if end_seconds <= start_seconds:
                continue

            include_hours = int(end_seconds // 3600) > 0
            include_minutes = include_hours or int((end_seconds % 3600) // 60) > 0
            start_tc = self._seconds_to_compact_timecode(
                start_seconds,
                fps_fraction,
                include_hours=include_hours,
                include_minutes=include_minutes,
            )
            end_tc = self._seconds_to_compact_timecode(
                end_seconds,
                fps_fraction,
                include_hours=include_hours,
                include_minutes=include_minutes,
            )

            split_data = json.loads(json.dumps(source_data))
            split_data.pop("id", None)
            split_data["start"] = start_seconds
            split_data["end"] = end_seconds
            split_data["name"] = "{} ({} to {})".format(
                base_name,
                start_tc,
                end_tc,
            )

            split_file = File()
            split_file.id = None
            split_file.key = None
            split_file.type = "insert"
            split_file.data = split_data
            self._append_scene_tag(split_file)
            split_file.save()
            created += 1
        return created

    def _resolve_caption_text(self, saved_paths, text_outputs):
        srt_path = ""
        for path in saved_paths:
            if str(path).lower().endswith(".srt"):
                srt_path = path
                break
        if srt_path:
            try:
                with open(srt_path, "r", encoding="utf-8") as handle:
                    text = handle.read().strip()
                if text:
                    return text
            except Exception as ex:
                log.warning("Failed reading SRT file for file caption metadata: %s", ex)

        for value in text_outputs:
            text = str(value or "").strip()
            if "-->" in text:
                return text

        for value in text_outputs:
            text = str(value or "").strip()
            if text:
                return text

        return ""

    def _store_caption_on_file(self, source_file_id, caption_text):
        caption_text = str(caption_text or "").strip()
        if not caption_text:
            return False

        source_file_value = source_file_id
        file_obj = File.get(id=source_file_value)
        if file_obj is None:
            file_obj = File.get(id=str(source_file_value or ""))
        if file_obj is None:
            log.info("No source file found for caption metadata update (file_id=%s)", source_file_value)
            return False

        if not isinstance(file_obj.data, dict):
            file_obj.data = {}
        file_obj.data["caption"] = caption_text
        file_obj.save()
        self.win.FileUpdated.emit(str(file_obj.id))
        return True

    def _seconds_to_compact_timecode(self, seconds_value, fps_fraction, include_hours=False, include_minutes=False):
        fps_fraction = fps_fraction if isinstance(fps_fraction, Fraction) and fps_fraction > 0 else Fraction(30, 1)
        fps_float = float(fps_fraction)
        frame_number = int(round(max(0.0, float(seconds_value or 0.0)) * fps_float)) + 1
        t = time_parts.secondsToTime((frame_number - 1) / fps_float, fps_fraction.numerator, fps_fraction.denominator)
        hours = int(t.get("hour", 0))
        minutes = int(t.get("min", 0))
        secs = int(t.get("sec", 0))
        frames = int(t.get("frame", 0))
        if include_hours:
            return "{:02d}:{:02d}:{:02d};{:02d}".format(hours, minutes, secs, frames)
        if include_minutes:
            return "{:02d}:{:02d};{:02d}".format(minutes, secs, frames)
        return "{:02d};{:02d}".format(secs, frames)

    def _append_scene_tag(self, file_obj):
        tags_raw = str(file_obj.data.get("tags", "") or "").strip()
        if not tags_raw:
            file_obj.data["tags"] = "scene"
            return
        tags = [part.strip() for part in tags_raw.split(",") if part.strip()]
        if any(part.lower() == "scene" for part in tags):
            return
        tags.append("scene")
        file_obj.data["tags"] = ", ".join(tags)

    def _apply_scene_segment_metadata(self, source_file_id, saved_paths):
        source_file = File.get(id=source_file_id) if source_file_id else None
        base_name = "scene"
        fps_fraction = Fraction(30, 1)
        if source_file:
            source_path = str(source_file.data.get("path", "") or "")
            if source_path:
                base_name = os.path.splitext(os.path.basename(source_path))[0] or base_name
            fps_data = source_file.data.get("fps", {})
            try:
                num = int(fps_data.get("num", 30))
                den = int(fps_data.get("den", 1) or 1)
                if num > 0 and den > 0:
                    fps_fraction = Fraction(num, den)
            except (TypeError, ValueError, ZeroDivisionError):
                fps_fraction = Fraction(30, 1)

        imported_files = []
        for path in saved_paths:
            file_obj = File.get(path=path)
            if file_obj and str(file_obj.data.get("media_type", "")) == "video":
                imported_files.append(file_obj)
        if not imported_files:
            return 0

        running_start = 0.0
        updated = 0
        for file_obj in imported_files:
            duration = float(file_obj.data.get("duration") or 0.0)
            if duration <= 0:
                start_trim = float(file_obj.data.get("start") or 0.0)
                end_trim = float(file_obj.data.get("end") or 0.0)
                duration = max(0.0, end_trim - start_trim)
            running_end = running_start + max(0.0, duration)

            include_hours = int(running_end // 3600) > 0
            include_minutes = include_hours or int((running_end % 3600) // 60) > 0
            start_tc = self._seconds_to_compact_timecode(
                running_start, fps_fraction, include_hours=include_hours, include_minutes=include_minutes
            )
            end_tc = self._seconds_to_compact_timecode(
                running_end, fps_fraction, include_hours=include_hours, include_minutes=include_minutes
            )
            file_obj.data["name"] = "{} ({} to {})".format(base_name, start_tc, end_tc)
            self._append_scene_tag(file_obj)
            file_obj.save()
            self.win.FileUpdated.emit(str(file_obj.id))

            running_start = running_end
            updated += 1

        return updated

    def _comfy_output_ref_from_path(self, path_text):
        """Convert a Comfy output absolute/relative path into a /view-compatible output ref."""
        path_text = unquote(str(path_text or "").strip())
        if not path_text:
            return None
        normalized = path_text.replace("\\", "/")
        filename = os.path.basename(normalized)
        if not filename:
            return None

        subfolder = ""
        marker = "/output/"
        if marker in normalized:
            rel = normalized.split(marker, 1)[1].lstrip("/")
            rel_dir = os.path.dirname(rel).strip("/")
            subfolder = rel_dir
        elif normalized.startswith("output/"):
            rel = normalized[len("output/"):]
            rel_dir = os.path.dirname(rel).strip("/")
            subfolder = rel_dir
        else:
            if os.path.isabs(normalized):
                # Unknown absolute location outside Comfy output tree; fallback to basename only.
                return {
                    "filename": filename,
                    "subfolder": "",
                    "type": "output",
                }
            rel_dir = os.path.dirname(normalized).strip("/")
            if rel_dir and rel_dir != ".":
                subfolder = rel_dir

        return {
            "filename": filename,
            "subfolder": subfolder,
            "type": "output",
        }

    def _next_generation_name(self, requested_name):
        base = re.sub(r"[^A-Za-z0-9._-]+", "_", str(requested_name or "").strip()).strip("._")
        if not base:
            base = "generation"

        existing_names = set()
        for file_obj in File.filter():
            if not file_obj:
                continue
            display_name = str(file_obj.data.get("name") or os.path.basename(file_obj.data.get("path", "")) or "")
            if display_name:
                stem = os.path.splitext(display_name)[0]
                existing_names.add(stem.lower())

        if base.lower() not in existing_names:
            return base

        name_root = base
        m = re.match(r"^(.*?)(?:_gen(\d+))?$", base, re.IGNORECASE)
        if m:
            name_root = (m.group(1) or base).rstrip("_") or "generation"
        n = 1
        while True:
            candidate = "{}_gen{}".format(name_root, n)
            if candidate.lower() not in existing_names:
                return candidate
            n += 1

    def _next_available_path(self, path):
        if not os.path.exists(path):
            return path
        folder = os.path.dirname(path)
        stem, ext = os.path.splitext(os.path.basename(path))
        n = 2
        while True:
            candidate = os.path.join(folder, "{}_{}{}".format(stem, n, ext))
            if not os.path.exists(candidate):
                return candidate
            n += 1



import os
import sys
import types
import unittest
from unittest.mock import patch

from PyQt5.QtCore import QRect, QSize
from PyQt5.QtWidgets import QApplication
import openshot


PATH = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
if PATH not in sys.path:
    sys.path.append(PATH)

from qt_test_app import ensure_app_state, get_or_create_app


class DummySettings:
    def __init__(self):
        self.values = {}

    def get(self, key):
        return self.values.get(key, False)

    def set(self, key, value):
        self.values[key] = value


class DummyApp(QApplication):
    def __init__(self):
        super().__init__([])
        self.settings = DummySettings()


app, _owns_app = get_or_create_app(DummyApp)
ensure_app_state(app, DummySettings, extra_attrs={"window": types.SimpleNamespace()})

from windows.cutting import Cutting
from windows.region import SelectRegion
from windows.preview_thread import PlayerWorker
from windows.video_widget import VideoWidget


class DummySignal:
    def __init__(self):
        self.calls = 0

    def emit(self, *args):
        self.calls += 1


class DummyVideoPreview:
    def __init__(self, viewport):
        self._viewport = viewport

    def centeredViewport(self, width, height):
        _ = (width, height)
        return self._viewport

    def width(self):
        return self._viewport.width()

    def height(self):
        return self._viewport.height()


class DialogPreviewResizeTests(unittest.TestCase):
    def test_cutting_select_reader_data_uses_proxy_when_target_fits(self):
        proxy = {"path": "/proxy.mp4", "width": 1280, "height": 720}
        source = {"path": "/source.mp4", "width": 3840, "height": 2160}
        fake = types.SimpleNamespace(
            proxy_reader_data=proxy,
            source_reader_data=source,
            _reader_capacity=lambda data: Cutting._reader_capacity(None, data),
        )

        reader_data = Cutting._select_reader_data_for_size(fake, QSize(960, 540))

        self.assertEqual(reader_data, proxy)

    def test_cutting_select_reader_data_promotes_to_source_when_proxy_too_small(self):
        proxy = {"path": "/proxy.mp4", "width": 1280, "height": 720}
        source = {"path": "/source.mp4", "width": 3840, "height": 2160}
        fake = types.SimpleNamespace(
            proxy_reader_data=proxy,
            source_reader_data=source,
            _reader_capacity=lambda data: Cutting._reader_capacity(None, data),
        )

        reader_data = Cutting._select_reader_data_for_size(fake, QSize(1920, 1080))

        self.assertEqual(reader_data, proxy)

    def test_cutting_target_preview_max_size_caps_to_reader_resolution(self):
        fake = types.SimpleNamespace(
            videoPreview=DummyVideoPreview(QRect(0, 0, 2400, 1350)),
            devicePixelRatioF=lambda: 1.0,
            width=1280,
            height=720,
        )

        size = Cutting._target_preview_max_size(fake)

        self.assertEqual(size, QSize(1280, 720))

    def test_region_target_preview_max_size_scales_with_viewport(self):
        fake = types.SimpleNamespace(
            videoPreview=DummyVideoPreview(QRect(0, 0, 900, 500)),
            devicePixelRatioF=lambda: 1.0,
            width=1920,
            height=1080,
        )

        size = SelectRegion._target_preview_max_size(fake)

        self.assertEqual(size, QSize(900, 500))

    def test_cutting_target_preview_max_size_uses_project_floor_for_single_image_media(self):
        fake = types.SimpleNamespace(
            videoPreview=DummyVideoPreview(QRect(0, 0, 900, 500)),
            devicePixelRatioF=lambda: 1.0,
            width=64,
            height=64,
            reader_data={"media_type": "image"},
            file=types.SimpleNamespace(data={"media_type": "image"}),
        )
        fake_app = types.SimpleNamespace(project=types.SimpleNamespace(get=lambda key: {"width": 1920, "height": 1080}.get(key, 0)))

        with patch("windows.cutting.get_app", return_value=fake_app):
            size = Cutting._target_preview_max_size(fake)

        self.assertEqual(size, QSize(900, 500))

    def test_cutting_apply_dynamic_preview_max_size_refreshes_when_size_changes(self):
        calls = []
        timeline = types.SimpleNamespace(
            preview_width=320,
            preview_height=180,
            SetMaxSize=lambda w, h: calls.append(("set", w, h)),
            ClearAllCache=lambda deep: calls.append(("clear", deep)),
        )
        fake = types.SimpleNamespace(
            initialized=True,
            r=timeline,
            _target_preview_max_size=lambda: QSize(640, 360),
            _select_reader_data_for_size=lambda size: {"path": "/source.mp4"},
            reader_data={"path": "/source.mp4"},
            PauseSignal=DummySignal(),
            refreshFrameSignal=DummySignal(),
        )

        Cutting._apply_dynamic_preview_max_size(fake)

        self.assertEqual(fake.PauseSignal.calls, 1)
        self.assertEqual(fake.refreshFrameSignal.calls, 1)
        self.assertEqual(calls, [("set", 640, 360), ("clear", True)])

    def test_cutting_apply_dynamic_preview_max_size_resumes_playback_when_was_playing(self):
        calls = []
        timeline = types.SimpleNamespace(
            preview_width=320,
            preview_height=180,
            SetMaxSize=lambda w, h: calls.append(("set", w, h)),
            ClearAllCache=lambda deep: calls.append(("clear", deep)),
        )
        player = types.SimpleNamespace(
            Mode=lambda: openshot.PLAYBACK_PLAY,
            Speed=lambda: 1.0,
        )
        fake = types.SimpleNamespace(
            initialized=True,
            r=timeline,
            preview_thread=types.SimpleNamespace(player=player),
            _target_preview_max_size=lambda: QSize(640, 360),
            _select_reader_data_for_size=lambda size: {"path": "/source.mp4"},
            reader_data={"path": "/source.mp4"},
            PauseSignal=DummySignal(),
            refreshFrameSignal=DummySignal(),
        )
        play_calls = []
        fake.btnPlay_clicked = lambda force=None: play_calls.append(force)

        with patch("windows.cutting.QTimer.singleShot", side_effect=lambda delay, fn: fn()):
            Cutting._apply_dynamic_preview_max_size(fake)

        self.assertEqual(fake.PauseSignal.calls, 1)
        self.assertEqual(fake.refreshFrameSignal.calls, 1)
        self.assertEqual(play_calls, ["play"])

    def test_cutting_preview_ready_seeks_frame_one_without_preroll(self):
        seek_calls = []
        fake = types.SimpleNamespace(
            is_preview_mode=True,
            preview_thread=types.SimpleNamespace(Seek=lambda frame, start_preroll=True: seek_calls.append((frame, start_preroll))),
            _preview_autoplay_active=False,
            SeekSignal=DummySignal(),
        )

        Cutting._preview_ready(fake)

        self.assertEqual(seek_calls, [(1, False)])

    def test_player_worker_play_honors_pending_seek_preroll_flag(self):
        applied = []
        player = types.SimpleNamespace(Play=lambda: applied.append(("play",)))
        worker = types.SimpleNamespace(
            parent=types.SimpleNamespace(initialized=True),
            player=player,
            _take_pending_seek=lambda: (1, False),
            _apply_seek=lambda frame, start_preroll: applied.append((frame, start_preroll)),
        )

        PlayerWorker.Play(worker)

        self.assertEqual(applied, [(1, False), ("play",)])

    def test_video_widget_resize_event_skips_pause_for_dialog_preview(self):
        pause_calls = []
        timer_calls = []
        fake = types.SimpleNamespace(
            delayed_size=None,
            size=lambda: QSize(640, 360),
            delayed_resize_timer=types.SimpleNamespace(start=lambda: timer_calls.append("start")),
            watch_project=False,
            win=types.SimpleNamespace(PauseSignal=types.SimpleNamespace(emit=lambda: pause_calls.append("pause"))),
        )
        event = types.SimpleNamespace(accept=lambda: None)

        VideoWidget.resizeEvent(fake, event)

        self.assertEqual(timer_calls, ["start"])
        self.assertEqual(pause_calls, [])

    def test_cutting_build_preview_timeline_uses_source_dimensions(self):
        timeline_args = []
        timeline_setmax = []

        class FakeTimeline:
            def __init__(self, width, height, fps, sample_rate, channels, channel_layout):
                timeline_args.append((width, height, fps, sample_rate, channels, channel_layout))
                self.info = types.SimpleNamespace()

            def SetMaxSize(self, width, height):
                timeline_setmax.append((width, height))

            def AddClip(self, clip):
                self.clip = clip

            def Open(self):
                pass

        class FakeClipReaderInfo:
            has_video = True
            has_audio = False

        class FakeClipReader:
            info = FakeClipReaderInfo()

        class FakeClip:
            def __init__(self, path):
                self.path = path
                self.reader = FakeClipReader()
                self.display = None

            def SetJson(self, payload):
                self.payload = payload

            def Start(self, value):
                self.start = value

            def End(self, value):
                self.end = value

            def Reader(self):
                return self.reader

        fake = types.SimpleNamespace(
            width=3840,
            height=2160,
            fps_num=30,
            fps_den=1,
            sample_rate=48000,
            channels=2,
            channel_layout=3,
            file=types.SimpleNamespace(absolute_path=lambda: "/source.mp4", data={"start": 0.0, "duration": 5.0}),
            video_length=150,
            source_reader_data={"path": "/source.mp4"},
            proxy_reader_data={"path": ""},
        )

        with patch("windows.cutting.openshot.Timeline", FakeTimeline), \
             patch("windows.cutting.openshot.Clip", FakeClip), \
             patch("windows.cutting.openshot.Fraction", side_effect=lambda num, den: (num, den)), \
             patch("windows.cutting.openshot.FRAME_DISPLAY_CLIP", 7):
            Cutting._build_preview_timeline(fake, {"path": "/source.mp4"}, QSize(640, 360))

        self.assertEqual(timeline_args[0][0:2], (3840, 2160))
        self.assertEqual(timeline_setmax, [(640, 360)])

    def test_cutting_build_preview_timeline_uses_project_floor_for_single_image_media(self):
        timeline_args = []
        clip_payloads = []

        class FakeTimeline:
            def __init__(self, width, height, fps, sample_rate, channels, channel_layout):
                timeline_args.append((width, height, fps, sample_rate, channels, channel_layout))
                self.info = types.SimpleNamespace()

            def SetMaxSize(self, width, height):
                self.max_size = (width, height)

            def AddClip(self, clip):
                self.clip = clip

            def Open(self):
                pass

        class FakeClipReaderInfo:
            has_video = True
            has_audio = False

        class FakeClipReader:
            info = FakeClipReaderInfo()

        class FakeClip:
            def __init__(self, path):
                self.path = path
                self.reader = FakeClipReader()
                self.display = None

            def SetJson(self, payload):
                clip_payloads.append(payload)
                self.payload = payload

            def Start(self, value):
                self.start = value

            def End(self, value):
                self.end = value

            def Reader(self):
                return self.reader

        fake = types.SimpleNamespace(
            width=64,
            height=64,
            fps_num=30,
            fps_den=1,
            sample_rate=48000,
            channels=2,
            channel_layout=3,
            file=types.SimpleNamespace(
                absolute_path=lambda: "/emoji.svg",
                data={"start": 0.0, "duration": 5.0, "media_type": "image"},
            ),
            video_length=150,
            source_reader_data={"path": "/emoji.svg", "media_type": "image"},
            proxy_reader_data={"path": ""},
        )
        fake_app = types.SimpleNamespace(project=types.SimpleNamespace(get=lambda key: {"width": 1920, "height": 1080}.get(key, 0)))

        with patch("windows.cutting.openshot.Timeline", FakeTimeline), \
             patch("windows.cutting.openshot.Clip", FakeClip), \
             patch("windows.cutting.openshot.Fraction", side_effect=lambda num, den: (num, den)), \
             patch("windows.cutting.openshot.FRAME_DISPLAY_CLIP", 7), \
             patch("windows.cutting.get_app", return_value=fake_app):
            Cutting._build_preview_timeline(fake, {"path": "/emoji.svg", "media_type": "image"}, QSize(640, 360))

        self.assertEqual(timeline_args[0][0:2], (1920, 1080))
        self.assertEqual(clip_payloads, [])


if __name__ == "__main__":
    unittest.main()



"""
 @file
 @brief Unit tests for Qt compatibility helpers
"""

import os
import sys
import unittest


PATH = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
if PATH not in sys.path:
    sys.path.append(PATH)

from classes import qt_types


class _OldIndex:
    def __init__(self):
        self.calls = []

    def row(self):
        return 7

    def sibling(self, row, column):
        self.calls.append((row, column))
        return ("sibling", row, column)


class _NewIndex(_OldIndex):
    def siblingAtColumn(self, column):
        self.calls.append(("at", column))
        return ("siblingAtColumn", column)


class _OldMetrics:
    def width(self, text):
        return len(text) * 10


class _NewMetrics(_OldMetrics):
    def horizontalAdvance(self, text):
        return len(text) * 12


class QtTypesTests(unittest.TestCase):
    def test_model_index_sibling_at_column_uses_new_api_when_available(self):
        index = _NewIndex()

        result = qt_types.model_index_sibling_at_column(index, 3)

        self.assertEqual(result, ("siblingAtColumn", 3))
        self.assertEqual(index.calls, [("at", 3)])

    def test_model_index_sibling_at_column_falls_back_to_old_api(self):
        index = _OldIndex()

        result = qt_types.model_index_sibling_at_column(index, 2)

        self.assertEqual(result, ("sibling", 7, 2))
        self.assertEqual(index.calls, [(7, 2)])

    def test_font_metrics_horizontal_advance_uses_new_api_when_available(self):
        metrics = _NewMetrics()

        self.assertEqual(qt_types.font_metrics_horizontal_advance(metrics, "abc"), 36)

    def test_font_metrics_horizontal_advance_falls_back_to_old_api(self):
        metrics = _OldMetrics()

        self.assertEqual(qt_types.font_metrics_horizontal_advance(metrics, "abc"), 30)



"""
 @file
 @brief This file contains the transitions file treeview, used by the main window
 @author Jonathan Thomas <jonathan@openshot.org>

 @section LICENSE

 Copyright (c) 2008-2018 OpenShot Studios, LLC
 (http://www.openshotstudios.com). This file is part of
 OpenShot Video Editor (http://www.openshot.org), an open-source project
 dedicated to delivering high quality video editing and animation solutions
 to the world.

 OpenShot Video Editor is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 OpenShot Video Editor is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with OpenShot Library.  If not, see <http://www.gnu.org/licenses/>.
 """

from PyQt5.QtCore import Qt, QSize, QPoint, QRegExp
from PyQt5.QtGui import QDrag
from PyQt5.QtWidgets import QListView, QAbstractItemView

from classes import info
from classes.app import get_app
from classes.logger import log
from .menu import StyledContextMenu


class TransitionsListView(QListView):
    """ A QListView QWidget used on the main window """
    drag_item_size = QSize(48, 48)
    drag_item_center = QPoint(24, 24)

    def contextMenuEvent(self, event):
        event.accept()

        # Set context menu mode
        app = get_app()
        app.context_menu_object = "transitions"

        menu = StyledContextMenu(parent=self)
        menu.addAction(self.win.actionDetailsView)
        menu.popup(event.globalPos())

    def startDrag(self, supportedActions):
        """ Override startDrag method to display custom icon """

        # Get first column indexes for all selected rows
        selected = self.selectionModel().selectedRows(0)

        # Get image of current item
        current = self.selectionModel().currentIndex()
        if not current.isValid() and selected:
            current = selected[0]

        if not current.isValid():
            log.warning("No draggable items found in model!")
            return False

        # Get icon from column 0 on same row as current item
        icon = current.sibling(current.row(), 0).data(Qt.DecorationRole)

        # Start drag operation
        drag = QDrag(self)
        drag.setMimeData(self.model().mimeData(selected))
        drag.setPixmap(icon.pixmap(self.drag_item_size))
        drag.setHotSpot(self.drag_item_center)
        drag.exec_()

    def filter_changed(self):
        self.refresh_view()

    def refresh_view(self):
        """Filter transitions with proxy class"""
        filter_text = self.win.transitionsFilter.text()
        # Apply filter to the source proxy model (not the single-column wrapper)
        self.transition_model.proxy_model.setFilterRegExp(QRegExp(filter_text.replace(' ', '.*')))
        self.transition_model.proxy_model.setFilterCaseSensitivity(Qt.CaseInsensitive)
        self.transition_model.proxy_model.sort(Qt.AscendingOrder)

    def __init__(self, model):
        # Invoke parent init
        QListView.__init__(self)

        # Get a reference to the window object
        app = get_app()
        self.win = app.window

        # Get Model data
        self.transition_model = model

        # Keep track of mouse press start position to determine when to start drag
        self.setAcceptDrops(True)
        self.setDragEnabled(True)
        self.setDropIndicatorShown(True)

        self.setModel(self.transition_model.list_proxy_model)

        # Remove the default selection model and wire up to the list-specific one
        self.selectionModel().deleteLater()
        self.setSelectionMode(QAbstractItemView.SingleSelection)
        self.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.setSelectionModel(self.transition_model.list_selection_model)

        # Setup header columns
        self.setIconSize(info.LIST_ICON_SIZE)
        self.setGridSize(info.LIST_GRID_SIZE)
        self.setViewMode(QListView.IconMode)
        self.setResizeMode(QListView.Adjust)
        self.setUniformItemSizes(True)
        self.setWordWrap(False)
        self.setTextElideMode(Qt.ElideRight)

        # setup filter events
        app.window.transitionsFilter.textChanged.connect(self.filter_changed)
        app.window.refreshTransitionsSignal.connect(self.refresh_view)



"""
 @file
 @brief This file contains the preview thread, used for displaying previews of the timeline
 @author Jonathan Thomas <jonathan@openshot.org>

 @section LICENSE

 Copyright (c) 2008-2018 OpenShot Studios, LLC
 (http://www.openshotstudios.com). This file is part of
 OpenShot Video Editor (http://www.openshot.org), an open-source project
 dedicated to delivering high quality video editing and animation solutions
 to the world.

 OpenShot Video Editor is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 OpenShot Video Editor is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with OpenShot Library.  If not, see <http://www.gnu.org/licenses/>.
 """

import time
import threading
import sip
import math

from PyQt5.QtCore import QObject, QThread, QTimer, pyqtSlot, pyqtSignal, QCoreApplication
from PyQt5.QtWidgets import QMessageBox
import openshot  # Python module for libopenshot (required video editing module installed separately)

from classes.app import get_app
from classes.logger import log
from classes.updates import UpdateInterface


class PreviewParent(QObject, UpdateInterface):
    """ Class which communicates with the PlayerWorker Class (running on a separate thread) """

    def changed(self, action):
        """ This method is invoked by the UpdateManager each time a change happens (i.e UpdateInterface) """

        # Ignore changes that don't affect libopenshot
        if action and len(action.key) >= 1 and action.key[0].lower() in ["files", "history", "markers", "layers", "scale", "profile", "sample_rate", "export_settings"]:
            return

        if not getattr(self, "_use_timeline_sync_max", True):
            return

        try:
            # Keep track of max timeline frame # on any updates to the timeline.
            self.timeline_max_length = max(1, int(get_app().window.timeline_sync.GetLastFrame() or 1))
            log.debug(f"Max timeline length/frames detected: {self.timeline_max_length}")
        except Exception as e:
            log.info("Error calculating max timeline length on PreviewParent: %s. %s" % (e, action.json(is_array=True)))

    # Signal when the frame position changes in the preview player
    def onPositionChanged(self, current_frame):
        timeline_last_frame = max(1, int(getattr(self, "timeline_max_length", 1) or 1))
        self.parent.movePlayhead(max(1, int(current_frame)))

        # Check if we are at the end of the timeline
        if self.worker.player.Mode() == openshot.PLAYBACK_PLAY:
            loop_preview = bool(getattr(self.parent, "loop_playback", False))
            if self.worker.player.Speed() > 0.0 and current_frame >= timeline_last_frame:
                if loop_preview:
                    # Loop preview playback back to the beginning.
                    self.worker.Seek(1)
                else:
                    # Yes, pause the video
                    self.parent.PauseSignal.emit()
                    # If the player got past the end of the project, go back.
                    self.worker.Seek(timeline_last_frame)
            if self.worker.player.Speed() < 0.0 and current_frame <= 1:
                if loop_preview:
                    # Loop rewind to the end frame.
                    self.worker.Seek(timeline_last_frame)
                else:
                    # If rewinding, and the player got past the first frame,
                    # pause and go to frame 1
                    self.parent.PauseSignal.emit()
                    self.worker.Seek(1)

    @pyqtSlot(int)
    @pyqtSlot(int, bool)
    def QueueSeek(self, frame, start_preroll=True):
        """Queue latest seek request for worker loop (non-blocking)."""
        self.worker.queue_seek(frame, start_preroll)

    # Signal when the playback mode changes in the preview player (i.e PLAY, PAUSE, STOP)
    def onModeChanged(self, current_mode):
        try:
            if current_mode == openshot.PLAYBACK_PLAY:
                self.parent.SetPlayheadFollow(False)
            else:
                self.parent.SetPlayheadFollow(True)
        except AttributeError:
            # Parent object doesn't need the playhead follow code
            pass

    # Signal when the playback encounters an error
    def onError(self, error):
        # Get translation object
        _ = get_app()._tr

        # Only JUCE audio errors bubble up here now
        QMessageBox.warning(self.parent, _("Audio Error"), _("Please fix the following error and restart OpenShot\n%s") % error)

    def Stop(self, wait_for_thread=True):
        """Disconnect preview parent from update manager and stop worker thread"""
        get_app().updates.disconnect_listener(self)

        # Stop preview thread
        self.worker.Stop()
        self.worker.kill()
        if self.background.isRunning():
            log.info("Stopping preview thread (running=%s)", self.background.isRunning())
        self.background.quit()
        if wait_for_thread:
            if not self.background.wait(5000):
                log.warning("Preview thread did not stop within 5 seconds")

    @pyqtSlot(object, object)
    def Init(self, parent, timeline, video_widget, max_length=1):
        # Important vars
        self.parent = parent
        self.timeline = timeline
        try:
            max_len = max(1, int(max_length))
        except (TypeError, ValueError):
            max_len = 1

        # Dialog previews (Split File / Region) pass their own max_length and
        # should not be clamped by the main timeline's last frame.
        self._use_timeline_sync_max = max_len <= 1
        if self._use_timeline_sync_max and getattr(get_app().window, "timeline_sync", None):
            self.timeline_max_length = max(1, int(get_app().window.timeline_sync.GetLastFrame() or 1))
        else:
            self.timeline_max_length = max_len

        # Background Worker Thread (for preview video process)
        self.background = QThread(self)
        self.background.setObjectName("preview_background")
        self.worker = PlayerWorker()  # no parent!

        # Init worker variables
        self.worker.Init(parent, timeline, video_widget)

        # Hook up signals to Background Worker
        self.worker.position_changed.connect(self.onPositionChanged)
        self.worker.mode_changed.connect(self.onModeChanged)
        if hasattr(self.parent, "_preview_ready"):
            self.worker.ready.connect(self.parent._preview_ready)
        if hasattr(self.parent, "_preview_mode_changed"):
            self.worker.mode_changed.connect(self.parent._preview_mode_changed)
        self.background.started.connect(self.worker.Start)
        self.worker.finished.connect(self.background.quit)
        self.worker.error_found.connect(self.onError)

        # Connect preview thread to main UI signals
        self.parent.previewFrameSignal.connect(self.worker.previewFrame)
        self.parent.refreshFrameSignal.connect(self.worker.refreshFrame)
        self.parent.LoadFileSignal.connect(self.worker.LoadFile)
        if hasattr(self.parent, "LoadFilePreviewSignal"):
            self.parent.LoadFilePreviewSignal.connect(self.worker.LoadFilePreview)
        self.parent.PlaySignal.connect(self.worker.Play)
        self.parent.PauseSignal.connect(self.worker.Pause)
        self.parent.SeekSignal.connect(self.QueueSeek)
        self.parent.LoadTimelineAndSeekSignal.connect(self.worker.LoadTimelineAndSeek)
        self.parent.SpeedSignal.connect(self.worker.Speed)
        self.parent.StopSignal.connect(self.worker.Stop)

        # Move Worker to new thread, and Start
        self.worker.moveToThread(self.background)
        self.background.start()

        # Add preview parent as listener for updates
        get_app().updates.add_listener(self)


class PlayerWorker(QObject):
    """ QT Player Worker Object (to preview video on a separate thread) """

    position_changed = pyqtSignal(int)
    mode_changed = pyqtSignal(object)
    error_found = pyqtSignal(object)
    ready = pyqtSignal()
    finished = pyqtSignal()

    @pyqtSlot(object, object)
    def Init(self, parent, timeline, videoPreview):
        self.parent = parent
        self.timeline = timeline
        self.videoPreview = videoPreview
        self.clip_path = None
        self.clip_reader = None
        self.original_speed = 0
        self.original_position = 0
        self.previous_clips = []
        self.previous_clip_readers = []
        self.is_running = True
        self.number = None
        self.current_frame = None
        self.current_mode = None
        self.reader_mode = "timeline"
        self.preview_stretch = False
        self._seek_lock = threading.Lock()
        self._pending_seek = None
        self._last_queued_seek_request = None
        self._last_applied_seek_request = None
        self._last_applied_seek_time = 0.0

        # Create QtPlayer class from libopenshot
        self.player = openshot.QtPlayer()

    def CheckAudioDevice(self):
        """Check if any audio devices initialization errors, default sample rate, and current open audio device"""
        s = get_app().get_settings()
        project = get_app().project

        def update_project_sample_rate_without_dirty(value):
            """Normalize startup audio settings without changing project dirty state."""
            previous_dirty = project.has_unsaved_changes
            get_app().updates.update_untracked(["sample_rate"], value)
            project.has_unsaved_changes = previous_dirty

        # Check audio init error
        audio_error = self.player.GetError()
        if audio_error:
            log.warning('Audio initialization error: %s', audio_error)
            self.error_found.emit(audio_error)

        # Check active sample rate from audio device
        # Parse string as float ("48000.0" -> 48000   OR   NaN)
        detected_sample_rate = float(self.player.GetDefaultSampleRate())
        if detected_sample_rate and not math.isnan(detected_sample_rate) and detected_sample_rate > 0.0:
            # Convert float to Integer
            detected_sample_rate_int = round(detected_sample_rate)

            settings_sample_rate = int(s.get("default-samplerate") or 48000)
            if detected_sample_rate_int != settings_sample_rate:
                log.warning("Your sample rate (%d) does not match OpenShot (%d). "
                            "Adjusting your 'Preferences->Preview->Default Sample Rate to match your "
                            "system rate: %d." % (detected_sample_rate_int,
                                                  settings_sample_rate,
                                                  detected_sample_rate_int))

                # Update default sample rate in settings
                s.set("default-samplerate", detected_sample_rate_int)

                # Update current project's sample rate, so we don't have some crazy
                # audio drift due to mis-matching sample rates
                update_project_sample_rate_without_dirty(detected_sample_rate_int)

        # Convert float 'settings' sample rate to Integer, if detected
        if type(s.get("default-samplerate")) == float:
            s.set("default-samplerate", detected_sample_rate_int)

        # Convert float 'project' sample rate to Integer, if detected
        if type(project.get("sample_rate")) == float:
            update_project_sample_rate_without_dirty(round(project.get("sample_rate")))

        # Check active audio device name and type from audio device
        active_audio_device = self.player.GetCurrentAudioDevice()
        audio_device_value = f"{active_audio_device.get_name()}||{active_audio_device.get_type()}"
        if s.get("playback-audio-device") != audio_device_value:
            log.warning("Your active audio device (%s) does not match OpenShot (%s). "
                        "Adjusting your 'Preferences->Playback->Audio Device' to match your "
                        "active audio device: %s" % (audio_device_value,
                                                     s.get("playback-audio-device"),
                                                     audio_device_value))
            s.set("playback-audio-device", audio_device_value)

            # Set libopenshot settings
            lib_settings = openshot.Settings.Instance()
            lib_settings.PLAYBACK_AUDIO_DEVICE_NAME = active_audio_device.get_name()
            lib_settings.PLAYBACK_AUDIO_DEVICE_TYPE = active_audio_device.get_type()

    @pyqtSlot()
    def Start(self):
        """ This method starts the video player """
        log.info("QThread Start Method Invoked")

        # Init new player
        self.initPlayer()

        # Connect player to timeline reader
        self.player.Reader(self.timeline)
        self.player.Play()
        self.player.Pause()
        self.ready.emit()

        # Check for any Player initialization errors (only JUCE errors bubble up here now)
        # But slightly delay, to allow for correct audio thread initialization with the
        # correct number of channels and sample rate
        QTimer.singleShot(1000, self.CheckAudioDevice)

        # Main loop, waiting for frames to process
        while self.is_running:
            seek_request = self._take_pending_seek()
            if seek_request is not None:
                seek_frame, start_preroll = seek_request
                self._apply_seek(seek_frame, start_preroll)

            # Emit position changed signal (if needed)
            if self.current_frame != self.player.Position():
                self.current_frame = self.player.Position()

                if not self.clip_path:
                    # Emit position of overall timeline (don't emit this for clip previews)
                    self.position_changed.emit(self.current_frame)

                    # TODO: Remove this hack and really determine what's blocking the main thread
                    # Try and keep things responsive
                    QCoreApplication.processEvents()

            # Emit mode changed signal (if needed)
            if self.player.Mode() != self.current_mode:
                self.current_mode = self.player.Mode()
                self.mode_changed.emit(self.current_mode)

            # wait for a small delay
            time.sleep(0.005)
            QCoreApplication.processEvents()

        self.finished.emit()
        log.debug('exiting playback thread')

    @pyqtSlot()
    def initPlayer(self):
        log.debug("initPlayer")

        # Get the address of the player's renderer (a QObject that emits signals when frames are ready)
        self.renderer_address = self.player.GetRendererQObject()
        self.player.SetQWidget(sip.unwrapinstance(self.videoPreview))
        self.renderer = sip.wrapinstance(self.renderer_address, QObject)
        self.videoPreview.connectSignals(self.renderer)

    def kill(self):
        """ Kill this thread """
        self.is_running = False

    def previewFrame(self, number):
        """ Preview a certain frame """
        # Mark frame number for processing
        self.Seek(number)

    def refreshFrame(self):
        """ Refresh a certain frame """
        # Selection/UI refresh signals can arrive during active playback.
        # Avoid seeking while playing, which can perturb frame progression.
        if self.player.Mode() == openshot.PLAYBACK_PLAY and self.player.Speed() != 0.0:
            return

        # Always load back in the timeline reader
        self.parent.LoadFileSignal.emit('')

        # Refreshes should not trigger preroll/cache invalidation behavior.
        refresh_frame = int(self.player.Position())
        self.Seek(refresh_frame, False)

    def LoadFile(self, path=None):
        """ Load a media file into the video player """
        self.LoadFilePreview(path, False)

    @pyqtSlot(str, bool)
    def LoadFilePreview(self, path=None, stretch=False):
        """Load a media file into the video player with optional stretch scaling."""
        # Check to see if this path is already loaded
        if path == self.clip_path:
            if self.reader_mode == "clip" and self.preview_stretch == bool(stretch):
                return
            if self.clip_reader:
                self.original_position = self.player.Position()
                self.player.Reader(self.clip_reader)
                self.reader_mode = "clip"
                self.preview_stretch = bool(stretch)
                self.Seek(1)
            return
        if not path and not self.clip_path and self.reader_mode == "timeline":
            return

        log.info("LoadFile %s" % path)

        # Determine the current frame of the timeline (when switching to a clip)
        seek_position = 1
        if path and not self.clip_path:
            # Track the current frame
            self.original_position = self.player.Position()

        # If blank path, switch back to self.timeline reader
        if not path:
            # Return to self.timeline reader
            log.debug("Set timeline reader again in player: %s" % self.timeline)
            self.player.Reader(self.timeline)
            self.reader_mode = "timeline"
            self.preview_stretch = False

            # Clear clip reader reference
            self.clip_reader = None
            self.clip_path = None

            # Switch back to last timeline position
            seek_position = self.original_position
        else:
            # Create new timeline reader (to preview selected clip)
            project = get_app().project

            # Get some settings from the project
            fps = project.get("fps")
            width = int(project.get("width"))
            height = int(project.get("height"))
            sample_rate = int(project.get("sample_rate"))
            channels = int(project.get("channels"))
            channel_layout = int(project.get("channel_layout"))
            timeline_sync = getattr(get_app().window, "timeline_sync", None)
            preview_width = getattr(getattr(timeline_sync, "timeline", None), "preview_width", 0)
            preview_height = getattr(getattr(timeline_sync, "timeline", None), "preview_height", 0)

            # Create an instance of a libopenshot Timeline object
            self.clip_reader = openshot.Timeline(width, height,
                                                 openshot.Fraction(fps["num"], fps["den"]),
                                                 sample_rate, channels, channel_layout)
            self.clip_reader.info.channel_layout = channel_layout
            self.clip_reader.info.has_audio = True
            self.clip_reader.info.has_video = True
            self.clip_reader.info.video_length = 999999
            self.clip_reader.info.duration = 999999
            self.clip_reader.info.sample_rate = sample_rate
            self.clip_reader.info.channels = channels
            if preview_width and preview_height:
                self.clip_reader.SetMaxSize(int(preview_width), int(preview_height))

            try:
                # Add clip for current preview file
                new_clip = openshot.Clip(path)
                try:
                    if new_clip.Reader().info.has_video:
                        self.clip_reader.info.has_audio = False
                        new_clip.Reader().info.has_audio = False
                except Exception:
                    log.debug("Failed to check has_video on clip reader for %s", path)
                if stretch:
                    new_clip.scale = openshot.SCALE_STRETCH
                    new_clip.gravity = openshot.GRAVITY_CENTER
                self.clip_reader.AddClip(new_clip)
            except:
                log.warning('Failed to load media file into video player: %s' % path)

            # Assign new clip_reader
            self.clip_path = path
            self.reader_mode = "clip"
            self.preview_stretch = bool(stretch)

            # Keep track of previous clip readers (so we can Close it later)
            self.previous_clips.append(new_clip)
            self.previous_clip_readers.append(self.clip_reader)

            # Open and set reader
            self.clip_reader.Open()
            self.player.Reader(self.clip_reader)

        # Close and destroy old clip readers (leaving the 3 most recent)
        while len(self.previous_clip_readers) > 3:
            log.debug('Removing old clips from preview: %s' % self.previous_clip_readers[0])
            previous_clip = self.previous_clips.pop(0)
            previous_clip.Close()
            previous_reader = self.previous_clip_readers.pop(0)
            previous_reader.Close()

        # Seek to frame 1, and resume speed
        if not path:
            QTimer.singleShot(0, lambda: self.Seek(seek_position))
        else:
            self.Seek(seek_position)

    def Play(self):
        """ Start playing the video player """

        # Start playback
        if self.parent.initialized:
            pending = self._take_pending_seek()
            if pending is not None:
                seek_frame, start_preroll = pending
                self._apply_seek(seek_frame, start_preroll)
            self.player.Play()

    def Pause(self):
        """ Pause the video player """

        # Pause playback
        if self.parent.initialized:
            self.player.Pause()

    def Stop(self):
        """ Stop the video player and terminate the playback threads """

        # Stop playback
        if self.parent.initialized:
            self.player.Stop()

    def queue_seek(self, number, start_preroll=True):
        """Queue latest seek request (latest-wins, thread-safe)."""
        if not self.parent.initialized:
            return
        frame = max(1, int(number))
        seek_request = (frame, bool(start_preroll))

        now = time.monotonic()

        # Drop immediate duplicate seeks (same frame + preroll) that can be
        # generated by UI signal fan-out during scrubbing/commit boundaries.
        if (
            self._last_applied_seek_request == seek_request
            and (now - self._last_applied_seek_time) < 0.03
        ):
            return
        with self._seek_lock:
            if self._last_queued_seek_request == seek_request:
                return
            self._last_queued_seek_request = seek_request
            self._pending_seek = seek_request

    def _take_pending_seek(self):
        with self._seek_lock:
            seek_request = self._pending_seek
            self._pending_seek = None
            if seek_request is not None:
                # Reset dedup key once the queued seek is consumed so future
                # refreshes at the same frame can trigger another render.
                self._last_queued_seek_request = None
            return seek_request

    def _apply_seek(self, frame, start_preroll):
        try:
            self.player.Seek(frame, start_preroll)
        except TypeError:
            # Backward compatibility with older libopenshot builds exposing
            # only Seek(frame).
            self.player.Seek(frame)
        self._last_applied_seek_request = (int(max(1, frame)), bool(start_preroll))
        self._last_applied_seek_time = time.monotonic()
        # Force the main loop to publish a fresh position_changed event after
        # each seek so timeline playheads stay in sync with queued seeks.
        self.current_frame = None

    @pyqtSlot(int)
    @pyqtSlot(int, bool)
    def Seek(self, number, start_preroll=True):
        """Seek to a specific frame (queued for worker loop application)."""
        self.queue_seek(number, start_preroll)

    @pyqtSlot(int)
    def LoadTimelineAndSeek(self, frame):
        frame = max(1, int(frame))
        self.original_position = frame
        if self.timeline:
            self.player.Reader(self.timeline)
            self.reader_mode = "timeline"
            self.clip_reader = None
            self.clip_path = None
        self.Seek(frame)

    def Speed(self, new_speed):
        """ Set the speed of the video player """

        # Set speed
        if self.parent.initialized and self.player.Speed() != new_speed:
            self.player.Speed(new_speed)



"""
 @file
 @brief This file is used to generate an EDL (edit decision list) export
 @author Jonathan Thomas <jonathan@openshot.org>

 @section LICENSE

 Copyright (c) 2008-2018 OpenShot Studios, LLC
 (http://www.openshotstudios.com). This file is part of
 OpenShot Video Editor (http://www.openshot.org), an open-source project
 dedicated to delivering high quality video editing and animation solutions
 to the world.

 OpenShot Video Editor is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 OpenShot Video Editor is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with OpenShot Library.  If not, see <http://www.gnu.org/licenses/>.
 """

import os
from operator import itemgetter

from PyQt5.QtWidgets import QFileDialog

from classes import info
from classes.app import get_app
from classes.logger import log
from classes.path_utils import relative_export_path, absolute_media_path
from classes.query import Clip, Track, File
from classes.time_parts import secondsToTimecode

def _interp_name(value):
    """Map interpolation value to a stable name."""
    try:
        numeric = int(value)
    except (TypeError, ValueError):
        numeric = None
    # libopenshot: 0=bezier, 1=linear, 2=constant/hold
    if numeric == 0:
        return "bezier"
    if numeric == 1:
        return "linear"
    if numeric == 2:
        return "hold"
    text = str(value).lower() if value is not None else ""
    if text.startswith("bez"):
        return "bezier"
    if text.startswith("hold") or text.startswith("const"):
        return "hold"
    return "linear"

def _is_drop_frame(fps_num, fps_den):
    """Return True if FPS corresponds to a common drop-frame rate."""
    if fps_den == 0:
        return False
    fps = float(fps_num) / float(fps_den)
    return any(abs(fps - rate) < 0.01 for rate in (29.97, 59.94))


def _clip_media_path(clip):
    file_id = clip.data.get("file_id")
    if file_id:
        file_obj = File.get(id=file_id)
        if file_obj:
            path = file_obj.absolute_path()
            if path:
                return path

    reader_path = clip.data.get("reader", {}).get("path")
    if reader_path:
        resolved = absolute_media_path(reader_path)
        if resolved:
            return resolved
        return reader_path
    return ""


def _volume_to_db(linear_value):
    """Convert linear (0-1) volume to dB, clamped to a reasonable floor."""
    try:
        v = float(linear_value)
    except (TypeError, ValueError):
        v = 0.0
    v = max(0.0, v)
    if v == 0.0:
        return -96.0
    import math
    db = 20.0 * math.log10(v)
    return max(db, -96.0)


def _fmt_value(value):
    """Format numeric with up to 2 decimals, stripping trailing zeros."""
    try:
        val = float(value)
    except (TypeError, ValueError):
        return "0"
    s = f"{val:.2f}"
    s = s.rstrip("0").rstrip(".")
    return s or "0"


def _fmt_percent(value):
    """Format percent by rounding to nearest int (no decimals)."""
    try:
        val = float(value)
    except (TypeError, ValueError):
        return "0"
    return str(int(round(val)))


def _db_to_volume(db_value):
    """Convert dB back to linear 0-1."""
    import math
    try:
        db = float(db_value)
    except (TypeError, ValueError):
        return 0.0
    if db <= -96.0:
        return 0.0
    linear = 10 ** (db / 20.0)
    return max(0.0, min(1.0, linear))


def export_edl():
    """Export EDL File"""
    app = get_app()
    _ = app._tr

    # EDL Export format
    edl_string = "%03d  %-9s%-6s%-9s%11s %11s %11s %11s\n"

    # Get FPS info
    fps_num = get_app().project.get("fps").get("num", 24)
    fps_den = get_app().project.get("fps").get("den", 1)
    fps_float = float(fps_num / fps_den)

    # Get EDL path
    recommended_path = app.project.current_filepath or ""
    if not recommended_path:
        recommended_path = os.path.join(info.HOME_PATH, "%s.edl" % _("Untitled Project"))
    else:
        recommended_path = recommended_path.replace(".osp", ".edl")
    file_path = QFileDialog.getSaveFileName(app.window, _("Export EDL..."), recommended_path,
                                            _("Edit Decision List (*.edl)"))[0]
    if not file_path:
        return

    # Append .edl if needed
    if not file_path.endswith(".edl"):
        file_path = "%s.edl" % file_path

    export_root = os.path.dirname(os.path.abspath(file_path))

    # Get filename with no extension
    file_name_with_ext = os.path.basename(file_path)
    file_name = os.path.splitext(file_name_with_ext)[0]

    all_tracks = get_app().project.get("layers")
    track_count = len(all_tracks)
    for track in reversed(sorted(all_tracks, key=itemgetter('number'))):
        existing_track = Track.get(number=track.get("number"))
        if not existing_track:
            # Log error and fail silently, and continue
            log.error('No track object found with number: %s' % track.get("number"))
            continue

        # Track name
        track_name = track.get("label") or "TRACK %s" % track_count
        clips_on_track = sorted(Clip.filter(layer=track.get("number")), key=lambda c: c.data.get('position', 0.0))
        if not clips_on_track:
            continue

        # Generate EDL File (1 per track - limitation of EDL format)
        # TODO: Improve and move this into its own class
        with open("%s-%s.edl" % (file_path.replace(".edl", ""), track_name), 'w', encoding="utf8") as f:
            # Add Header
            f.write("TITLE: %s - %s\n" % (file_name, track_name))
            f.write("FCM: %s\n\n" % ("DROP FRAME" if _is_drop_frame(fps_num, fps_den) else "NON-DROP FRAME"))

            # Loop through each track
            export_position = 0.0
            event_index = 1

            # Loop through clips on this track
            for clip in clips_on_track:
                clip_position = clip.data.get('position', 0.0)
                clip_start = clip.data.get('start', 0.0)
                clip_end = clip.data.get('end', clip_start)
                clip_duration = clip_end - clip_start

                # Do we need a blank clip?
                if clip_position > export_position:
                    clip_start_time = secondsToTimecode(0.0, fps_num, fps_den)
                    clip_end_time = secondsToTimecode(clip_position - export_position, fps_num, fps_den)
                    timeline_start_time = secondsToTimecode(export_position, fps_num, fps_den)
                    timeline_end_time = secondsToTimecode(clip_position, fps_num, fps_den)

                    f.write(edl_string % (
                        event_index, "BL"[:9], "V"[:6], "C",
                        clip_start_time, clip_end_time,
                        timeline_start_time, timeline_end_time))
                    event_index += 1
                    export_position = clip_position

                # Format clip start/end and timeline start/end values (i.e. 00:00:00:00)
                clip_start_time = secondsToTimecode(clip_start, fps_num, fps_den)
                clip_end_time = secondsToTimecode(clip_end, fps_num, fps_den)
                timeline_start_time = secondsToTimecode(clip_position, fps_num, fps_den)
                timeline_end_time = secondsToTimecode(clip_position + clip_duration, fps_num, fps_den)

                has_video = clip.data.get("reader", {}).get("has_video", False)
                has_audio = clip.data.get("reader", {}).get("has_audio", False)
                if not has_video and not has_audio:
                    continue
                reel_name = clip.data.get("reel") or "AX"
                reel_video_tag = f"{reel_name} V"
                reel_audio_tag = f"{reel_name} A1"
                if has_video:
                    # Video Track
                    f.write(edl_string % (
                            event_index, reel_name[:9], "V"[:6], "C",
                            clip_start_time, clip_end_time,
                            timeline_start_time, timeline_end_time))
                if has_audio:
                    # Audio Track
                    f.write(edl_string % (
                            event_index, reel_name[:9], "A"[:6], "C",
                        clip_start_time, clip_end_time,
                        timeline_start_time, timeline_end_time))
                f.write("* FROM CLIP NAME: %s\n" % clip.data.get('title'))
                media_path = _clip_media_path(clip)
                relative_media = relative_export_path(media_path, export_root)
                if relative_media:
                    f.write("* SOURCE FILE: %s\n" % relative_media)

                # Add opacity data (if any)
                alpha_points = clip.data.get('alpha', {}).get('Points', [])
                if len(alpha_points) >= 1:
                    # Loop through Points (remove duplicates)
                    keyframes = {}
                    for point in alpha_points:
                        keyframeTime = (point.get('co', {}).get('X', 1.0) - 1) / fps_float
                        keyframeValue = point.get('co', {}).get('Y', 0.0) * 100.0
                        interp_name = _interp_name(point.get("interpolation"))
                        keyframes[keyframeTime] = (keyframeValue, interp_name)
                    # Write keyframe values to EDL
                    for opacity_time in sorted(keyframes.keys()):
                        opacity_value, interp_name = keyframes.get(opacity_time)
                        tc = secondsToTimecode(opacity_time, fps_num, fps_den)
                        f.write("* VIDEO LEVEL AT %s IS %s%% %s (REEL %s)\n" % (tc, _fmt_percent(opacity_value), interp_name.upper(), reel_video_tag))

                # Add volume data (if any)
                volume_points = clip.data.get('volume', {}).get('Points', [])
                if len(volume_points) >= 1:
                    # Loop through Points (remove duplicates)
                    keyframes = {}
                    for point in volume_points:
                        keyframeTime = (point.get('co', {}).get('X', 1.0) - 1) / fps_float
                        keyframeValue = _volume_to_db(point.get('co', {}).get('Y', 0.0))
                        interp_name = _interp_name(point.get("interpolation"))
                        keyframes[keyframeTime] = (keyframeValue, interp_name)
                    # Write keyframe values to EDL
                    for volume_time in sorted(keyframes.keys()):
                        volume_value, interp_name = keyframes.get(volume_time)
                        f.write("* AUDIO LEVEL AT %s IS %.2f DB %s (REEL %s)\n" % (secondsToTimecode(volume_time, fps_num, fps_den), volume_value, interp_name.upper(), reel_audio_tag))

                # Export transform keyframes (skip defaults)
                transform_defs = [
                    ("scale_x", "SCALE X", 100.0, 1.0, True),
                    ("scale_y", "SCALE Y", 100.0, 1.0, True),
                    ("location_x", "LOCATION X", 100.0, 0.0, True),
                    ("location_y", "LOCATION Y", 100.0, 0.0, True),
                    ("rotation", "ROTATION", 1.0, 0.0, False),
                    ("shear_x", "SHEAR X", 100.0, 0.0, True),
                    ("shear_y", "SHEAR Y", 100.0, 0.0, True),
                ]

                for key_name, label, multiplier, default_val, is_percent in transform_defs:
                    points = clip.data.get(key_name, {}).get("Points", [])
                    if not points:
                        continue
                    include_all = len(points) > 1
                    keyframes = {}
                    for point in points:
                        keyframeTime = (point.get('co', {}).get('X', 1.0) - 1) / fps_float
                        raw_value = point.get('co', {}).get('Y', default_val)
                        if not include_all and len(points) == 1 and abs(raw_value - default_val) < 1e-6:
                            continue  # single default point: skip
                        if not include_all and keyframes and abs(raw_value - default_val) < 1e-6:
                            continue  # avoid duplicate defaults when not including all
                        display_value = raw_value * multiplier if is_percent else raw_value
                        interp_name = _interp_name(point.get("interpolation"))
                        keyframes[keyframeTime] = (display_value, interp_name)
                    for t in sorted(keyframes.keys()):
                        val, interp_name = keyframes[t]
                        unit = "%" if is_percent else "DEG"
                        display_val = _fmt_percent(val) if is_percent else _fmt_value(val)
                        if unit == "%":
                            value_unit = "%s%%" % display_val
                        else:
                            value_unit = "%s %s" % (display_val, unit.strip())
                        f.write("* %s AT %s IS %s %s (REEL %s)\n" % (label, secondsToTimecode(t, fps_num, fps_den), value_unit, interp_name.upper(), reel_video_tag))

                # Update export position
                export_position = max(export_position, clip_position + clip_duration)
                event_index += 1
                f.write("\n")

            # Update counters
            track_count -= 1



"""
 @file
 @brief Painter for track backgrounds and labels.
 @author Jonathan Thomas <jonathan@openshot.org>

 @section LICENSE

 Copyright (c) 2008-2025 OpenShot Studios, LLC
 (http://www.openshotstudios.com). This file is part of
 OpenShot Video Editor (http://www.openshot.org), an open-source project
 dedicated to delivering high quality video editing and animation solutions
 to the world.

 OpenShot Video Editor is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 OpenShot Video Editor is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with OpenShot Library.  If not, see <http://www.gnu.org/licenses/>.
 """

from PyQt5.QtCore import QPointF, QRectF, Qt
import math

from PyQt5.QtGui import (
    QBrush,
    QColor,
    QLinearGradient,
    QPainter,
    QPainterPath,
    QPen,
)

from .base import BasePainter


class TrackPainter(BasePainter):
    def update_theme(self):
        self.border_pen = QPen(self.w.theme.track.border_color)
        self.border_pen.setCosmetic(True)
        self.name_border_color = self.w.theme.track.name_border_color
        self.name_border_width = self.w.theme.track.name_border_width
        self.name_border_top_color = self.w.theme.track.name_border_top_color
        self.name_border_top_width = self.w.theme.track.name_border_top_width
        self.name_border_bottom_color = self.w.theme.track.name_border_bottom_color
        self.name_border_bottom_width = self.w.theme.track.name_border_bottom_width
        self.name_radius_tl = self.w.theme.track.name_radius_tl
        self.name_radius_bl = self.w.theme.track.name_radius_bl
        self.name_top_overlay = QColor(self.w.theme.track.name_top_overlay)
        self.name_top_overlay2 = QColor(self.w.theme.track.name_top_overlay2)
        self.menu_pix = None
        if self.w.theme.menu_icon:
            size = self.w.theme.menu_size or self.w.theme.menu_icon.width()
            self.menu_pix = self.scaled_pixmap(self.w.theme.menu_icon, size, size)
        self.menu_margin = self.w.theme.menu_margin
        self.toggle_off_pix = None
        self.toggle_on_pix = None
        toggle_size = float(self.w.theme.menu_size or 0.0)

        def _scaled_toggle(pixmap):
            if not pixmap or pixmap.isNull():
                return None
            width = float(pixmap.width())
            height = float(pixmap.height())
            if toggle_size > 0.0:
                target = max(toggle_size, width, height)
                width = height = target
            return self.scaled_pixmap(pixmap, width, height)

        if self.w.theme.keyframe_toggle_off_icon:
            self.toggle_off_pix = _scaled_toggle(
                self.w.theme.keyframe_toggle_off_icon
            )
        if self.w.theme.keyframe_toggle_on_icon:
            self.toggle_on_pix = _scaled_toggle(
                self.w.theme.keyframe_toggle_on_icon
            )
        self.toggle_margin = self.w.theme.menu_margin

        self.toolbar_order = (
            "lock-toggle",
            "keyframe-panel",
        )

        toolbar = {}

        keyframe_disabled = _scaled_toggle(
            getattr(self.w.theme, "track_keyframe_panel_disabled_icon", None)
            or self.w.theme.keyframe_toggle_off_icon
        )
        keyframe_enabled = _scaled_toggle(
            getattr(self.w.theme, "track_keyframe_panel_enabled_icon", None)
            or self.w.theme.keyframe_toggle_on_icon
        )
        if keyframe_disabled or keyframe_enabled:
            toolbar["keyframe-panel"] = {
                "disabled": keyframe_disabled,
                "enabled": keyframe_enabled or keyframe_disabled,
            }
            if not self.toggle_off_pix:
                self.toggle_off_pix = keyframe_disabled
            if not self.toggle_on_pix:
                self.toggle_on_pix = keyframe_enabled or keyframe_disabled

        lock_locked_disabled = _scaled_toggle(getattr(self.w.theme, "track_locked_disabled_icon", None))
        lock_locked_enabled = _scaled_toggle(getattr(self.w.theme, "track_locked_enabled_icon", None))
        lock_unlocked_disabled = _scaled_toggle(getattr(self.w.theme, "track_unlocked_disabled_icon", None))
        lock_unlocked_enabled = _scaled_toggle(getattr(self.w.theme, "track_unlocked_enabled_icon", None))
        if (
            lock_locked_disabled
            or lock_locked_enabled
            or lock_unlocked_disabled
            or lock_unlocked_enabled
        ):
            toolbar["lock-toggle"] = {
                "locked": {
                    "disabled": lock_locked_disabled,
                    "enabled": lock_locked_enabled or lock_locked_disabled,
                },
                "unlocked": {
                    "disabled": lock_unlocked_disabled,
                    "enabled": lock_unlocked_enabled or lock_unlocked_disabled,
                },
            }

        self.toolbar_pixmaps = toolbar

    def _track_name_path(self, rect: QRectF) -> QPainterPath:
        r = QRectF(rect)
        radius_tl = max(0.0, float(self.name_radius_tl or 0.0))
        radius_bl = max(0.0, float(self.name_radius_bl or 0.0))
        radius_tl = min(radius_tl, r.height() / 2.0)
        radius_bl = min(radius_bl, r.height() / 2.0)
        path = QPainterPath()
        if radius_tl <= 0.0 and radius_bl <= 0.0:
            path.addRect(r)
            return path
        path.moveTo(r.x() + radius_tl, r.y())
        path.lineTo(r.right(), r.y())
        path.lineTo(r.right(), r.bottom())
        path.lineTo(r.x() + radius_bl, r.bottom())
        if radius_bl > 0.0:
            path.quadTo(r.x(), r.bottom(), r.x(), r.bottom() - radius_bl)
        else:
            path.lineTo(r.x(), r.bottom())
        if radius_tl > 0.0:
            path.lineTo(r.x(), r.y() + radius_tl)
            path.quadTo(r.x(), r.y(), r.x() + radius_tl, r.y())
        else:
            path.lineTo(r.x(), r.y())
        path.closeSubpath()
        return path

    def paint_background(self, painter: QPainter):
        area = QRectF(
            self.w.track_name_width,
            self.w.ruler_height,
            self.w.width() - self.w.track_name_width - self.w.scroll_bar_thickness,
            self.w.height() - self.w.ruler_height - self.w.scroll_bar_thickness,
        )
        painter.save()
        painter.setClipRect(area)
        banding_cfg = self._frame_banding_config()
        for track_rect, track, _name_rect in self.w.geometry.iter_tracks():
            vis = track_rect.intersected(area)
            if vis.isNull():
                continue
            locked = bool((track.data if isinstance(track.data, dict) else {}).get("lock"))
            bg = QColor(self.w.theme.track.background)
            bg2 = QColor(self.w.theme.track.background2)
            border_color = QColor(self.w.theme.track.border_color)
            if locked:
                bg = self.dimmed_color(bg)
                if bg2.isValid():
                    bg2 = self.dimmed_color(bg2)
                border_color = self.dimmed_color(border_color)
            if bg2.isValid() and bg2 != bg:
                grad = QLinearGradient(vis.topLeft(), vis.bottomLeft())
                grad.setColorAt(0, bg)
                grad.setColorAt(1, bg2)
                painter.fillRect(vis, QBrush(grad))
            else:
                painter.fillRect(vis, bg)
            if banding_cfg:
                per_track_cfg = banding_cfg
                if locked:
                    per_track_cfg = dict(banding_cfg)
                    per_track_cfg["color"] = self.dimmed_color(QColor(banding_cfg.get("color")))
                self._paint_frame_banding(painter, vis, per_track_cfg)
            border_pen = QPen(border_color)
            border_pen.setCosmetic(True)
            painter.setPen(border_pen)
            painter.drawLine(vis.topLeft(), vis.topRight())
            painter.drawLine(vis.bottomLeft(), vis.bottomRight())
            painter.drawLine(vis.topRight(), vis.bottomRight())

        painter.fillRect(
            self.w.resize_handle_rect.intersected(area),
            self.w.theme.track.border_color,
        )
        timeline_handle = self.w.geometry.timeline_handle_rect()
        if timeline_handle and not timeline_handle.isNull():
            handle_rect = timeline_handle.intersected(area)
            if not handle_rect.isNull():
                painter.fillRect(handle_rect, self.w.theme.track.border_color)
                inner = QRectF(handle_rect)
                inner.adjust(1.0, 1.0, -1.0, -1.0)
                if inner.width() > 0 and inner.height() > 0:
                    accent = QColor(self.w.theme.track.name_background)
                    accent.setAlpha(180)
                    painter.fillRect(inner, accent)
        painter.restore()

    def _frame_banding_config(self):
        pps = float(getattr(self.w, "pixels_per_second", 0.0) or 0.0)
        fps = float(getattr(self.w, "fps_float", 0.0) or 0.0)
        if pps <= 0.0 or fps <= 0.0:
            return None
        ruler = getattr(self.w, "ruler_painter", None)
        if not ruler:
            return None
        try:
            fpt = ruler._frames_per_tick(pps, fps)
        except Exception:
            return None
        if fpt > 2.0:
            return None
        pixels_per_frame = pps / fps if fps else 0.0
        if pixels_per_frame <= 0.0:
            return None
        base_color = QColor(self.w.theme.track.background)
        if not base_color.isValid():
            return None
        if base_color.lightness() < 128:
            band_color = base_color.lighter(125)
        else:
            band_color = base_color.darker(110)
        band_color.setAlpha(min(220, max(120, band_color.alpha())))
        return {
            "pps": pps,
            "fps": fps,
            "fpt": fpt,
            "pixels_per_frame": pixels_per_frame,
            "offset_px": float(getattr(self.w, "h_scroll_offset", 0.0) or 0.0),
            "color": band_color,
        }

    def _paint_frame_banding(self, painter: QPainter, rect: QRectF, cfg):
        pixels_per_frame = cfg.get("pixels_per_frame", 0.0)
        if pixels_per_frame <= 0.0:
            return
        pps = cfg.get("pps", 0.0)
        fps = cfg.get("fps", 0.0)
        if pps <= 0.0 or fps <= 0.0:
            return
        offset_px = cfg.get("offset_px", 0.0)
        # Convert visible rect to absolute timeline pixel positions
        left_px = offset_px + max(0.0, rect.left() - self.w.track_name_width)
        right_px = offset_px + max(0.0, rect.right() - self.w.track_name_width)
        if right_px <= left_px:
            return
        start_seconds = left_px / pps
        end_seconds = right_px / pps
        if end_seconds <= start_seconds:
            return

        # Use a small epsilon when converting to frame numbers so that
        # fractional floating point rounding errors don't cause the
        # computed frame parity to jitter as we scroll horizontally.
        eps = 1e-9
        start_frame = int(math.floor(start_seconds * fps + eps))
        end_frame = int(math.ceil(end_seconds * fps - eps))
        if end_frame <= start_frame:
            end_frame = start_frame + 1
        painter.save()
        painter.setPen(Qt.NoPen)
        painter.setBrush(cfg.get("color"))
        # Ensure we do not draw excessive rectangles
        max_frames = end_frame - start_frame
        if max_frames > 2000:
            max_frames = 2000
            end_frame = start_frame + max_frames
        for frame in range(start_frame, end_frame):
            if frame % 2 != 0:
                continue
            t = frame / fps
            x = self.w.track_name_width + t * pps - offset_px
            band_rect = QRectF(x, rect.top(), pixels_per_frame, rect.height())
            band_rect = band_rect.intersected(rect)
            if band_rect.width() <= 0.0:
                continue
            painter.drawRect(band_rect)
        painter.restore()

    def paint_names(self, painter: QPainter):
        area = QRectF(
            0,
            self.w.ruler_height,
            self.w.track_name_width,
            self.w.height() - self.w.ruler_height - self.w.scroll_bar_thickness,
        )
        painter.save()
        painter.setClipRect(area)
        for _track_rect, track, name_rect in self.w.geometry.iter_tracks():
            locked = bool((track.data if isinstance(track.data, dict) else {}).get("lock"))
            name_bg = QColor(self.w.theme.track.name_background)
            name_border_top = QColor(self.name_border_top_color)
            name_border_bottom = QColor(self.name_border_bottom_color)
            name_border = QColor(self.name_border_color)
            text_color = QColor(self.w.theme.track.font_color)
            if locked:
                name_bg = self.dimmed_color(name_bg)
                name_border_top = self.dimmed_color(name_border_top)
                name_border_bottom = self.dimmed_color(name_border_bottom)
                name_border = self.dimmed_color(name_border)
                text_color = self.dimmed_color(text_color)
            painter.setPen(Qt.NoPen)
            painter.setBrush(name_bg)
            path = self._track_name_path(name_rect)
            painter.drawPath(path)

            # Match JS .track_top overlay (light-to-transparent).
            overlay_top = QColor(self.name_top_overlay)
            overlay_bottom = QColor(self.name_top_overlay2)
            if overlay_top.isValid() or overlay_bottom.isValid():
                if not overlay_top.isValid() and overlay_bottom.isValid():
                    overlay_top = QColor(overlay_bottom)
                if not overlay_bottom.isValid() and overlay_top.isValid():
                    overlay_bottom = QColor(overlay_top)
                    overlay_bottom.setAlpha(0)
                if locked:
                    overlay_top = self.dimmed_color(overlay_top)
                    overlay_bottom = self.dimmed_color(overlay_bottom)
                grad = QLinearGradient(name_rect.topLeft(), name_rect.bottomLeft())
                grad.setColorAt(0.0, overlay_top)
                grad.setColorAt(1.0, overlay_bottom)
                painter.save()
                painter.setClipPath(path)
                painter.fillRect(name_rect, QBrush(grad))
                painter.restore()
            painter.setBrush(Qt.NoBrush)

            painter.save()
            painter.setClipPath(path)
            radius_tl = max(0.0, float(self.name_radius_tl or 0.0))
            radius_bl = max(0.0, float(self.name_radius_bl or 0.0))
            if self.name_border_top_width:
                top_rect = QRectF(
                    name_rect.x() + radius_tl,
                    name_rect.y(),
                    max(0.0, name_rect.width() - radius_tl),
                    self.name_border_top_width,
                )
                painter.fillRect(top_rect, name_border_top)
            if self.name_border_bottom_width:
                bottom_rect = QRectF(
                    name_rect.x() + radius_bl,
                    name_rect.bottom() - self.name_border_bottom_width,
                    max(0.0, name_rect.width() - radius_bl),
                    self.name_border_bottom_width,
                )
                painter.fillRect(bottom_rect, name_border_bottom)
            if self.name_border_width:
                left_rect = QRectF(
                    name_rect.x(),
                    name_rect.y() + radius_tl,
                    self.name_border_width,
                    max(
                        0.0,
                        name_rect.height()
                        - radius_tl
                        - radius_bl,
                    ),
                )
                painter.fillRect(left_rect, name_border)

            # Preserve curved left-corner border strokes on rounded track names.
            if radius_tl > 0.0 and self.name_border_top_width and name_border_top.isValid():
                pen = QPen(name_border_top, float(self.name_border_top_width))
                pen.setCapStyle(Qt.FlatCap)
                pen.setJoinStyle(Qt.RoundJoin)
                painter.setPen(pen)
                painter.setBrush(Qt.NoBrush)
                arc = QRectF(
                    name_rect.x(),
                    name_rect.y(),
                    radius_tl * 2.0,
                    radius_tl * 2.0,
                )
                painter.drawArc(arc, 90 * 16, 90 * 16)
            if radius_bl > 0.0 and self.name_border_bottom_width and name_border_bottom.isValid():
                pen = QPen(name_border_bottom, float(self.name_border_bottom_width))
                pen.setCapStyle(Qt.FlatCap)
                pen.setJoinStyle(Qt.RoundJoin)
                painter.setPen(pen)
                painter.setBrush(Qt.NoBrush)
                arc = QRectF(
                    name_rect.x(),
                    name_rect.bottom() - (radius_bl * 2.0),
                    radius_bl * 2.0,
                    radius_bl * 2.0,
                )
                painter.drawArc(arc, 180 * 16, 90 * 16)
            painter.restore()

            menu_w = 0.0
            metrics = painter.fontMetrics()
            text_height = float(metrics.height()) if metrics else 0.0
            text_top = name_rect.y() + self.menu_margin
            text_bottom_limit = name_rect.bottom() - self.menu_margin
            if text_height > 0.0 and text_top + text_height > text_bottom_limit:
                text_height = max(0.0, text_bottom_limit - text_top)

            if self.menu_pix:
                menu_x = name_rect.x() + self.name_border_width + self.menu_margin
                menu_w, menu_h = self.logical_size(self.menu_pix)
                menu_y = text_top
                if text_height > 0.0:
                    menu_y += max(0.0, (text_height - menu_h) / 2.0)
                painter.drawPixmap(QPointF(menu_x, menu_y), self.menu_pix)

            buttons = self.w._track_toolbar_buttons(track, name_rect)
            text_offset = self.name_border_width + self.menu_margin * 2 + menu_w
            right_padding = self.name_border_width + self.menu_margin
            text_width = max(0.0, name_rect.width() - text_offset - right_padding)
            text_rect_height = text_height if text_height > 0.0 else max(0.0, name_rect.height() - self.menu_margin * 2)
            text_bottom = min(text_top + text_rect_height, text_bottom_limit)
            text_rect = QRectF(
                name_rect.x() + text_offset,
                text_top,
                text_width,
                max(0.0, text_bottom - text_top),
            )
            painter.setPen(text_color)
            if text_rect.width() > 0.0 and text_rect.height() > 0.0:
                painter.drawText(
                    text_rect,
                    Qt.AlignLeft | Qt.AlignTop,
                    self.w._track_display_label(track)
                )

            hover_key = getattr(self.w, "_toolbar_hover_key", None)
            pressed_key = getattr(self.w, "_toolbar_pressed_key", None)
            pressed_inside = getattr(self.w, "_toolbar_pressed_inside", False)
            for button in buttons:
                button_key = (button.get("track_id"), button.get("key"))
                pix = self.w._toolbar_button_pixmap(
                    track,
                    button,
                    hovered=hover_key == button_key,
                    pressed=pressed_key == button_key and pressed_inside,
                )
                if not pix:
                    continue
                painter.save()
                if locked:
                    painter.setOpacity(0.8)
                default_margin = float(getattr(self, "toggle_margin", 0.0) or 0.0)
                margin_x = button.get("margin_x", button.get("margin", default_margin))
                margin_y = button.get("margin_y", button.get("margin", default_margin))
                draw_x = button["rect"].x() + margin_x
                draw_y = button["rect"].y() + margin_y
                painter.drawPixmap(QPointF(draw_x, draw_y), pix)
                painter.restore()
        painter.restore()



"""
 @file
 @brief Shared helpers for timeline color mappings.
 @author Jonathan Thomas <jonathan@openshot.org>

 @section LICENSE

 Copyright (c) 2008-2025 OpenShot Studios, LLC
 (http://www.openshotstudios.com). This file is part of
 OpenShot Video Editor (http://www.openshot.org), an open-source project
 dedicated to delivering high quality video editing and animation solutions
 to the world.

 OpenShot Video Editor is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 OpenShot Video Editor is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with OpenShot Library.  If not, see <http://www.gnu.org/licenses/>.
 """

from typing import Any, Dict

from PyQt5.QtGui import QColor

# Mapping of known effect names to their representative colors. These values
# mirror the existing web timeline so both backends stay consistent.
_EFFECT_COLORS: Dict[str, str] = {
    "AnalogTape": "#907600",
    "Bars": "#4d7bff",
    "Blur": "#0095bf",
    "Brightness": "#5500ff",
    "Caption": "#5e7911",
    "ChromaKey": "#00ad2d",
    "Color Shift": "#b39373",
    "Compressor": "#A52A2A",
    "Crop": "#7b3f00",
    "Deinterlace": "#006001",
    "Delay": "#ff4dd4",
    "Distortion": "#7393B3",
    "Echo": "#5C4033",
    "Expander": "#C4A484",
    "Hue": "#2d7b6b",
    "Mask": "#cb0091",
    "Negate": "#ff9700",
    "Noise": "#a9a9a9",
    "ObjectDetection": "#636363",
    "Parametric EQ": "#708090",
    "Pixelate": "#9fa131",
    "Robotization": "#CC5500",
    "Saturation": "#ff3d00",
    "Sharpen": "#49759c",
    "ColorMap": "#4d945d",
    "LensFlare": "#7c29d1",
    "Normalize": "#607d3b",
    "Outline": "#be6d33",
    "SphericalProjection": "#b886ea",
    "Shift": "#8d7960",
    "Stabilizer": "#9F2B68",
    "Tracker": "#DE3163",
    "Wave": "#FF00Ff",
    "Whisperization": "#93914a",
}

_DEFAULT_COLOR = "#4d7bff"


def _effect_type_name(effect: Any) -> str:
    if isinstance(effect, dict):
        for key in ("type", "effect", "name", "class_name"):
            value = effect.get(key)
            if value:
                return str(value)
    elif effect:
        return str(effect)
    return ""


def effect_color_hex(effect: Any) -> str:
    """Return the preferred hex color string for *effect*."""
    if isinstance(effect, dict):
        color = effect.get("color")
        if isinstance(color, str) and color:
            return color
    type_name = _effect_type_name(effect)
    return _EFFECT_COLORS.get(type_name, _DEFAULT_COLOR)


def effect_color_qcolor(effect: Any) -> QColor:
    """Return a QColor matching the preferred color for *effect*."""
    return QColor(effect_color_hex(effect))



"""
 @file
 @brief This file contains the add to timeline file treeview
 @author Jonathan Thomas <jonathan@openshot.org>

 @section LICENSE

 Copyright (c) 2008-2018 OpenShot Studios, LLC
 (http://www.openshotstudios.com). This file is part of
 OpenShot Video Editor (http://www.openshot.org), an open-source project
 dedicated to delivering high quality video editing and animation solutions
 to the world.

 OpenShot Video Editor is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 OpenShot Video Editor is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with OpenShot Library.  If not, see <http://www.gnu.org/licenses/>.
 """

from PyQt5.QtCore import QSize
from PyQt5.QtWidgets import QTreeView, QAbstractItemView

from classes import info
from classes.app import get_app
from windows.models.add_to_timeline_model import TimelineModel


class TimelineTreeView(QTreeView):
    """ A TreeView QWidget used on the add to timeline window """

    def currentChanged(self, selected, deselected):
        # Get selected item
        self.selected = selected
        self.deselected = deselected

        # Get translation object
        _ = self.app._tr

    def contextMenuEvent(self, event):
        # # Ignore event, propagate to parent
        event.ignore()

    def mousePressEvent(self, event):

        # Ignore event, propagate to parent
        event.ignore()
        super().mousePressEvent(event)

    def refresh_view(self):
        self.timeline_model.update_model()
        self.hideColumn(2)

    def __init__(self, *args):
        # Invoke parent init
        QTreeView.__init__(self, *args)

        # Get a reference to the window object
        self.app = get_app()
        self.win = args[0]

        # Get Model data
        self.timeline_model = TimelineModel()

        # Keep track of mouse press start position to determine when to start drag
        self.selected = None
        self.deselected = None

        # Setup header columns
        self.setModel(self.timeline_model.model)
        self.setIconSize(info.TREE_ICON_SIZE)
        self.setIndentation(0)
        self.setSelectionBehavior(QTreeView.SelectRows)
        self.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.setWordWrap(True)

        # Refresh view
        self.refresh_view()
